import pulsar
from pulsar import asyncio

if pulsar.appengine:
    def start_store(url, workers=0, **kw):
        raise RuntimeError('Cannot start datastore in google appengine')

else:
    from pulsar import (when_monitor_start, coroutine_return, get_application,
                        send)
    from pulsar.apps.data import create_store
    from pulsar.apps.ds import PulsarDS

    def start_pulsar_ds(arbiter, host, workers=0):
        lock = getattr(arbiter, 'lock', None)
        if lock is None:
            arbiter.lock = lock = asyncio.Lock()
        yield lock.acquire()
        try:
            app = yield get_application('pulsards')
            if not app:
                app = PulsarDS(bind=host, workers=workers)
                cfg = yield app(arbiter)
            else:
                cfg = app.cfg
            coroutine_return(cfg)
        finally:
            lock.release()

    def start_store(url, workers=0, **kw):
        '''Equivalent to :func:`.create_store` for most cases excepts when the
        ``url`` is for a pulsar store not yet started.
        In this case, a :class:`.PulsarDS` is started.
        '''
        store = create_store(url, **kw)
        if store.name == 'pulsar':
            client = store.client()
            try:
                yield client.ping()
            except pulsar.ConnectionRefusedError:
                host = localhost(store._host)
                if not host:
                    raise
                cfg = yield send('arbiter', 'run', start_pulsar_ds,
                                 host, workers)
                store._host = cfg.addresses[0]
                dns = store._buildurl()
                store = create_store(dns, **kw)
        coroutine_return(store)

    def localhost(host):
        if isinstance(host, tuple):
            if host[0] in ('127.0.0.1', ''):
                return ':'.join((str(b) for b in host))
        else:
            return host

    def _start_store(monitor):
        app = monitor.app
        if not isinstance(app, PulsarDS):
            dns = app.cfg.data_store
            if dns:
                store = yield start_store(dns)
                app.cfg.set('data_store', store.dns)

    when_monitor_start.append(_start_store)



import base64
import hashlib
from functools import partial

from pulsar import HttpException, ProtocolError, ProtocolConsumer, maybe_async
from pulsar.utils.pep import to_bytes, native_str
from pulsar.utils.httpurl import DEFAULT_CHARSET
from pulsar.utils.websocket import frame_parser, parse_close
from pulsar.apps import wsgi

from . import extensions

WEBSOCKET_GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
TRANSPORTS = {}


def register_transport(klass):
    TRANSPORTS[klass.name] = klass
    return klass


@register_transport
class WebSocket(wsgi.Router):
    """A :ref:`Router <wsgi-router>` for a websocket handshake.

    Once the handshake is succesful, the protocol consumer
    is upgraded to :class:`WebSocketProtocol` and messages are handled by
    the :attr:`handle` attribute, an instance of :class:`.WS`.

    See http://tools.ietf.org/html/rfc6455 for the websocket server protocol
    and http://www.w3.org/TR/websockets/ for details on the JavaScript
    interface.

    .. attribute:: parser_factory

        A factory of websocket frame parsers
    """
    parser_factory = frame_parser
    _name = 'websocket'

    def __init__(self, route, handle, parser_factory=None, **kwargs):
        super(WebSocket, self).__init__(route, **kwargs)
        self.handle = handle
        self.parser_factory = parser_factory or frame_parser

    @property
    def name(self):
        return self._name

    def get(self, request):
        headers_parser = self.handle_handshake(request.environ)
        if not headers_parser:
            raise HttpException(status=404)
        headers, parser = headers_parser
        response = request.response
        response.status_code = 101
        response.content = b''
        response.headers.update(headers)
        connection = request.environ.get('pulsar.connection')
        if not connection:
            raise HttpException(status=404)
        factory = partial(WebSocketProtocol, request, self.handle, parser)
        connection.upgrade(factory)
        return request.response

    def handle_handshake(self, environ):
        connections = environ.get(
            "HTTP_CONNECTION", '').lower().replace(' ', '').split(',')
        if environ.get("HTTP_UPGRADE", '').lower() != "websocket" or \
           'upgrade' not in connections:
            raise HttpException(status=400)
        key = environ.get('HTTP_SEC_WEBSOCKET_KEY')
        if key:
            try:
                ws_key = base64.b64decode(key.encode(DEFAULT_CHARSET))
            except Exception:
                ws_key = ''
            if len(ws_key) != 16:
                raise HttpException(msg="WebSocket key's length is invalid",
                                    status=400)
        else:
            raise HttpException(msg='Not a valid HyBi WebSocket request. '
                                    'Missing Sec-Websocket-Key header.',
                                status=400)
        # Collect supported subprotocols
        subprotocols = environ.get('HTTP_SEC_WEBSOCKET_PROTOCOL')
        ws_protocols = []
        if subprotocols:
            for s in subprotocols.split(','):
                ws_protocols.append(s.strip())
        # Collect supported extensions
        ws_extensions = []
        extensions = environ.get('HTTP_SEC_WEBSOCKET_EXTENSIONS')
        if extensions:
            for ext in extensions.split(','):
                ws_extensions.append(ext.strip())
        # Build the frame parser
        version = environ.get('HTTP_SEC_WEBSOCKET_VERSION')
        try:
            parser = self.parser_factory(version=version,
                                         protocols=ws_protocols,
                                         extensions=ws_extensions)
        except ProtocolError as e:
            raise HttpException(str(e), status=400)
        headers = [('Sec-WebSocket-Accept', self.challenge_response(key))]
        if parser.protocols:
            headers.append(('Sec-WebSocket-Protocol',
                            ', '.join(parser.protocols)))
        if parser.extensions:
            headers.append(('Sec-WebSocket-Extensions',
                            ','.join(parser.extensions)))
        return headers, parser

    def challenge_response(self, key):
        sha1 = hashlib.sha1(to_bytes(key+WEBSOCKET_GUID))
        return native_str(base64.b64encode(sha1.digest()))


class WebSocketProtocol(ProtocolConsumer):
    '''A :class:`.ProtocolConsumer` for websocket servers and clients.

    .. attribute:: handshake

        The original handshake response/request.

    .. attribute:: handler

        A websocket handler :class:`.WS`.

    .. attribute:: parser

        A websocket :class:`.FrameParser`.

    .. attribute:: close_reason

        A tuple of (``code``, ``reason``) or ``None``.

        Available when a close frame is received.
    '''
    close_reason = None

    def __init__(self, handshake, handler, parser):
        super(WebSocketProtocol, self).__init__()
        self.bind_event('post_request', self._shut_down)
        self.handshake = handshake
        self.handler = handler
        self.parser = parser

    @property
    def cfg(self):
        '''The :class:`.Config` container for this protocol.
        '''
        return self.handshake.cfg

    def connection_made(self, connection):
        connection.set_timeout(0)
        maybe_async(self.handler.on_open(self), self._loop)

    def data_received(self, data):
        frame = self.parser.decode(data)
        while frame:
            if frame.is_close:
                try:
                    self.close_reason = parse_close(frame.body)
                finally:
                    self._connection.close()
                break
            if frame.is_message:
                maybe_async(self.handler.on_message(self, frame.body))
            elif frame.is_bytes:
                maybe_async(self.handler.on_bytes(self, frame.body))
            elif frame.is_ping:
                maybe_async(self.handler.on_ping(self, frame.body))
            elif frame.is_pong:
                maybe_async(self.handler.on_pong(self, frame.body))
            frame = self.parser.decode()

    def write(self, message, opcode=None, **kw):
        '''Write a new ``message`` into the wire.

        It uses the :meth:`~.FrameParser.encode` method of the
        websocket :attr:`parser`.

        :param message: message to send, must be a string or bytes
        :param opcode: optional ``opcode``, if not supplied it is set to 1
            if ``message`` is a string, otherwise ``2`` when the message
            are bytes.
         '''
        chunk = self.parser.encode(message, opcode=opcode, **kw)
        self.transport.write(chunk)
        if opcode == 8:
            self.finish()

    def ping(self, message=None):
        '''Write a ping ``frame``.
        '''
        self.transport.write(self.parser.ping(message))

    def pong(self, message=None):
        '''Write a pong ``frame``.
        '''
        self.transport.write(self.parser.pong(message))

    def write_close(self, code=None):
        '''Write a close ``frame`` with ``code``.
        '''
        self.transport.write(self.parser.close(code))
        self._connection.close()

    def _shut_down(self, result, exc=None):
        maybe_async(self.handler.on_close(self))



'''
Pulsar-ds is a python implementation of the popular redis_
data store. It uses pulsar asynchronous framework to create a
single-threaded workers responding to TCP-requests in the same way
as redis does.

To run a stand alone server create a script with the following code::


    from pulsar.apps.data import PulsarDS

    if __name__ == '__main__':
        PulsarDS().start()


More information on the :ref:`pulsar data store example <tutorials-pulsards>`.

Check out these benchmarks_

.. _benchmarks: https://gist.github.com/lsbardel/8068579

Implementation
===========================

Pulsar Data Store Server
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: PulsarDS
   :members:
   :member-order: bysource


Storage
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: Storage
   :members:
   :member-order: bysource


.. _redis: http://redis.io/
'''
import os
import re
import time
import math
from random import choice
from hashlib import sha1
from itertools import islice, chain
from functools import partial, reduce
from collections import namedtuple

import pulsar
from pulsar.apps.socket import SocketServer
from pulsar.utils.config import Global
from pulsar.utils.structures import Dict, Zset, Deque
from pulsar.utils.pep import map, range, zip, ispy3k, pickle
try:
    from pulsar.utils.lua import Lua
except ImportError:     # pragma    nocover
    Lua = None


from .parser import redis_parser
from .utils import sort_command, count_bytes, and_op, or_op, xor_op, save_data
from .client import (command, PulsarStoreClient, LuaClient, Blocked,
                     COMMANDS_INFO, check_input, redis_to_py_pattern)


DEFAULT_PULSAR_STORE_ADDRESS = '127.0.0.1:6410'


def pulsards_url(address=None, db=3):
    address = address or DEFAULT_PULSAR_STORE_ADDRESS
    return 'pulsar://%s/%s' % (address, db)


# Keyspace changes notification classes
STRING_LIMIT = 2**32

nan = float('nan')

if ispy3k:
    from itertools import zip_longest
    _ord = lambda x: x
else:   # pragma    nocover
    from itertools import izip_longest as zip_longest
    _ord = lambda x: ord(x)


class RedisParserSetting(Global):
    name = "redis_py_parser"
    flags = ["--redis-py-parser"]
    action = "store_true"
    default = False
    desc = '''\
    Use the python redis parser rather the C implementation.

    Mainly used for benchmarking purposes.
    '''


class PulsarDsSetting(pulsar.Setting):
    virtual = True
    app = 'pulsards'
    section = "Pulsar data store server"


def validate_list_of_pairs(val):
    new_val = []
    if val:
        if not isinstance(val, (list, tuple)):
            raise TypeError("Not a list: %s" % val)
        for elem in val:
            if not isinstance(elem, (list, tuple)):
                raise TypeError("Not a list: %s" % elem)
            if not len(elem) == 2:
                raise TypeError("Not a pair: %s" % str(elem))
            new_val.append((int(elem[0]), int(elem[1])))
    return new_val


# #############################################################################
# #    CONFIGURATION PARAMETERS
class KeyValueDatabases(PulsarDsSetting):
    name = "key_value_databases"
    flags = ["--key-value-databases"]
    type = int
    default = 16
    desc = 'Number of databases for the key value store.'


class KeyValuePassword(PulsarDsSetting):
    name = "key_value_password"
    flags = ["--key-value-password"]
    default = ''
    desc = 'Optional password for the database.'


class KeyValueSave(PulsarDsSetting):
    name = "key_value_save"
    default = [(900, 1), (300, 10), (60, 10000)]
    validator = validate_list_of_pairs
    desc = '''\
        List of pairs controlling data store persistence.

        Will save the DB if both the given number of seconds and the given
        number of write operations against the DB occurred.

        The default behaviour will be to save:
        after 900 sec (15 min) if at least 1 key changed
        after 300 sec (5 min) if at least 10 keys changed
        after 60 sec if at least 10000 keys changed

        You can disable saving at all by setting an empty list
    '''


class KeyValueFileName(PulsarDsSetting):
    name = "key_value_filename"
    flags = ["--key-value-filename"]
    default = 'pulsards.rdb'
    desc = '''The filename where to dump the DB.'''


class TcpServer(pulsar.TcpServer):

    def __init__(self, cfg, *args, **kwargs):
        super(TcpServer, self).__init__(*args, **kwargs)
        self.cfg = cfg
        self._parser_class = redis_parser(cfg.redis_py_parser)
        self._key_value_store = Storage(self, cfg)

    def info(self):
        info = super(TcpServer, self).info()
        info.update(self._key_value_store._info())
        return info


class PulsarDS(SocketServer):
    '''A :class:`.SocketServer` serving a pulsar datastore.
    '''
    name = 'pulsards'
    cfg = pulsar.Config(bind=DEFAULT_PULSAR_STORE_ADDRESS,
                        keep_alive=0,
                        apps=['socket', 'pulsards'])

    def server_factory(self, *args, **kw):
        return TcpServer(self.cfg, *args, **kw)

    def protocol_factory(self):
        return partial(PulsarStoreClient, self.cfg)

    def monitor_start(self, monitor):
        cfg = self.cfg
        workers = min(1, cfg.workers)
        cfg.set('workers', workers)
        return super(PulsarDS, self).monitor_start(monitor)


# #############################################################################
# #    DATA STORE
pubsub_patterns = namedtuple('pubsub_patterns', 're clients')


class Storage(object):
    '''Implement redis commands.
    '''
    def __init__(self, server, cfg):
        self.cfg = cfg
        self._password = cfg.key_value_password.encode('utf-8')
        self._filename = cfg.key_value_filename
        self._writer = None
        self._server = server
        self._loop = server._loop
        self._parser = server._parser_class()
        self._missed_keys = 0
        self._hit_keys = 0
        self._expired_keys = 0
        self._dirty = 0
        self._bpop_blocked_clients = 0
        self._last_save = int(time.time())
        self._channels = {}
        self._patterns = {}
        # The set of clients which are watching keys
        self._watching = set()
        # The set of clients which issued the monitor command
        self._monitors = set()
        self.logger = server.logger
        #
        self.NOTIFY_KEYSPACE = (1 << 0)
        self.NOTIFY_KEYEVENT = (1 << 1)
        self.NOTIFY_GENERIC = (1 << 2)
        self.NOTIFY_STRING = (1 << 3)
        self.NOTIFY_LIST = (1 << 4)
        self.NOTIFY_SET = (1 << 5)
        self.NOTIFY_HASH = (1 << 6)
        self.NOTIFY_ZSET = (1 << 7)
        self.NOTIFY_EXPIRED = (1 << 8)
        self.NOTIFY_EVICTED = (1 << 9)
        self.NOTIFY_ALL = (self.NOTIFY_GENERIC | self.NOTIFY_STRING |
                           self.NOTIFY_LIST | self.NOTIFY_SET |
                           self.NOTIFY_HASH | self.NOTIFY_ZSET |
                           self.NOTIFY_EXPIRED | self.NOTIFY_EVICTED)

        self.MONITOR = (1 << 2)
        self.MULTI = (1 << 3)
        self.BLOCKED = (1 << 4)
        self.DIRTY_CAS = (1 << 5)
        #
        self._event_handlers = {self.NOTIFY_GENERIC: self._generic_event,
                                self.NOTIFY_STRING: self._string_event,
                                self.NOTIFY_SET: self._set_event,
                                self.NOTIFY_HASH: self._hash_event,
                                self.NOTIFY_LIST: self._list_event,
                                self.NOTIFY_ZSET: self._zset_event}
        self._set_options = (b'ex', b'px', 'nx', b'xx')
        self.OK = b'+OK\r\n'
        self.QUEUED = b'+QUEUED\r\n'
        self.ZERO = b':0\r\n'
        self.ONE = b':1\r\n'
        self.NIL = b'$-1\r\n'
        self.NULL_ARRAY = b'*-1\r\n'
        self.INVALID_TIMEOUT = 'invalid expire time'
        self.PUBSUB_ONLY = ('only (P)SUBSCRIBE / (P)UNSUBSCRIBE / QUIT '
                            'allowed in this context')
        self.INVALID_SCORE = 'Invalid score value'
        self.NOT_SUPPORTED = 'Command not yet supported'
        self.OUT_OF_BOUND = 'Out of bound'
        self.SYNTAX_ERROR = 'Syntax error'
        self.SUBSCRIBE_COMMANDS = ('psubscribe', 'punsubscribe', 'subscribe',
                                   'unsubscribe', 'quit')
        self.encoder = pickle
        self.hash_type = Dict
        self.list_type = Deque
        self.zset_type = Zset
        self.data_types = (bytearray, set, self.hash_type,
                           self.list_type, self.zset_type)
        self.zset_aggregate = {b'min': min,
                               b'max': max,
                               b'sum': sum}
        self._type_event_map = {bytearray: self.NOTIFY_STRING,
                                self.hash_type: self.NOTIFY_HASH,
                                self.list_type: self.NOTIFY_LIST,
                                set: self.NOTIFY_SET,
                                self.zset_type: self.NOTIFY_ZSET}
        self._type_name_map = {bytearray: 'string',
                               self.hash_type: 'hash',
                               self.list_type: 'list',
                               set: 'set',
                               self.zset_type: 'zset'}
        self.databases = dict(((num, Db(num, self))
                               for num in range(cfg.key_value_databases)))
        # Initialise lua
        if Lua:
            self.lua = Lua()
            self.scripts = {}
            self.lua.register('redis', LuaClient(self),
                              'call', 'pcall', 'error_reply', 'status_reply')
            self.version = '2.6.16'
        else:   # pragma    nocover
            self.lua = None
            self.version = '2.4.10'
        self._loaddb()
        pulsar.call_repeatedly(self._loop, 1, self._cron)

    # #########################################################################
    # #    KEYS COMMANDS
    @command('Keys', True, name='del')
    def delete(self, client, request, N):
        check_input(request, not N)
        rem = client.db.rem
        result = reduce(lambda x, y: x + rem(y), request[1:], 0)
        client.reply_int(result)

    @command('Keys')
    def dump(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_bulk()
        else:
            client.reply_bulk(self.encoder.dumps(value))

    @command('Keys')
    def exists(self, client, request, N):
        check_input(request, N != 1)
        if client.db.exists(request[1]):
            client.reply_one()
        else:
            client.reply_zero()

    @command('Keys', True)
    def expire(self, client, request, N, m=1):
        check_input(request, N != 2)
        try:
            timeout = int(request[2])
        except ValueError:
            client.reply_error(self.INVALID_TIMEOUT)
        else:
            if timeout:
                if timeout < 0:
                    return client.reply_error(self.INVALID_TIMEOUT)
                if client.db.expire(request[1], m*timeout):
                    return client.reply_one()
            client.reply_zero()

    @command('Keys', True)
    def expireat(self, client, request, N, M=1):
        check_input(request, N != 2)
        try:
            timeout = int(request[2])
        except ValueError:
            client.reply_error(self.INVALID_TIMEOUT)
        else:
            if timeout:
                if timeout < 0:
                    return client.reply_error(self.INVALID_TIMEOUT)
                timeout = M*timeout - time.time()
                if client.db.expire(request[1], timeout):
                    return client.reply_one()
            client.reply_zero()

    @command('Keys')
    def keys(self, client, request, N):
        err = 'ignore'
        check_input(request, N != 1)
        pattern = request[1].decode('utf-8', err)
        allkeys = pattern == '*'
        gr = None
        if not allkeys:
            gr = re.compile(redis_to_py_pattern(pattern))
        result = [key for key in client.db if allkeys or
                  gr.search(key.decode('utf-8', err))]
        client.reply_multi_bulk(result)

    @command('Keys', supported=False)
    def migrate(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Keys', True)
    def move(self, client, request, N):
        check_input(request, N != 2)
        key = request[1]
        try:
            db2 = self.databases.get(int(request[2]))
            if db2 is None:
                raise ValueError
        except Exception:
            return client.reply_zero()
        db = client.db
        value = db.get(key)
        if db2.exists(key) or value is None:
            return client.reply_zero()
        assert value
        db.pop(key)
        self._signal(self.NOTIFY_GENERIC, db, 'del', key, 1)
        db2._data[key] = value
        self._signal(self._type_event_map[type(value)], db2, 'set', key, 1)
        client.reply_one()

    @command('Keys', supported=False)
    def object(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Keys', True)
    def persist(self, client, request, N):
        check_input(request, N != 1)
        if client.db.persist(request[1]):
            client.reply_one()
        else:
            client.reply_zero()

    @command('Keys', True)
    def pexpire(self, client, request, N):
        self.expire(client, request, N, 0.001)

    @command('Keys', True)
    def pexpireat(self, client, request, N, M=1):
        self.expireat(client, request, N, 0.001)

    @command('Keys')
    def pttl(self, client, request, N):
        check_input(request, N != 1)
        client.reply_int(client.db.ttl(request[1], 1000))

    @command('Keys')
    def randomkey(self, client, request, N):
        check_input(request, N)
        keys = list(client.db)
        if keys:
            client.reply_bulk(choice(keys))
        else:
            client.reply_bulk()

    @command('Keys', True)
    def rename(self, client, request, N, ex=False):
        check_input(request, N != 2)
        key1, key2 = request[1], request[2]
        db = client.db
        value = db.get(key1)
        if value is None:
            client.reply_error('Cannot rename key, not available')
        elif key1 == key2:
            client.reply_error('Cannot rename key')
        else:
            assert value
            if ex:
                if db.exists(key2):
                    return client.reply_zero()
                result = 1
            else:
                result = 0
                if db.pop(key2) is not None:
                    self._signal(self.NOTIFY_GENERIC, db, 'del', key2)
            db.pop(key1)
            event = self._type_event_map[type(value)]
            dirty = 1 if event == self.NOTIFY_STRING else len(value)
            db._data[key2] = value
            self._signal(event, db, request[0], key2, dirty)
            client.reply_one() if result else client.reply_ok()

    @command('Keys', True)
    def renamenx(self, client, request, N):
        self.rename(client, request, N, True)

    @command('Keys', True)
    def restore(self, client, request, N):
        check_input(request, N != 3)
        key = request[1]
        db = client.db
        try:
            value = self.encoder.loads(request[3])
        except Exception:
            value = None
        if not isinstance(value, self.data_types):
            return client.reply_error('Could not decode value')
        try:
            ttl = int(request[2])
        except Exception:
            return client.reply_error(self.INVALID_TIMEOUT)
        if db.pop(key) is not None:
            self._signal(self.NOTIFY_GENERIC, db, 'del', key)
        db._data[key] = value
        if ttl > 0:
            db.expire(key, ttl)
        client.reply_ok()

    @command('Keys', True)
    def sort(self, client, request, N):
        check_input(request, not N)
        value = client.db.get(request[1])
        if value is None:
            value = self.list_type()
        elif not isinstance(value, (set, self.list_type, self.zset_type)):
            return client.reply_wrongtype()
        sort_command(self, client, request, value)

    @command('Keys', True)
    def ttl(self, client, request, N):
        check_input(request, N != 1)
        client.reply_int(client.db.ttl(request[1]))

    @command('Keys', True)
    def type(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            result = 'none'
        else:
            result = self._type_name_map[type(value)]
        client.reply_status(result)

    @command('Keys', supported=False)
    def scan(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    # #########################################################################
    # #    STRING COMMANDS
    @command('Strings', True)
    def append(self, client, request, N):
        check_input(request, N != 2,)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            value = bytearray(request[2])
            db._data[key] = value
        elif not isinstance(value, bytearray):
            return client.reply_wrongtype()
        else:
            assert value
            value.extend(request[2])
        self._signal(self.NOTIFY_STRING, db, request[0], key, 1)
        client.reply_int(len(value))

    @command('Strings')
    def bitcount(self, client, request, N):
        check_input(request, N < 1 or N > 3)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_int(0)
        elif not isinstance(value, bytearray):
            return client.reply_wrongtype()
        else:
            assert value
            if N > 1:
                start = request[2]
                end = request[3] if N == 3 else -1
                start, end = self._range_values(value, start, end)
                value = value[start:end]
            client.reply_int(count_bytes(value))

    @command('Strings', True)
    def bitop(self, client, request, N):
        check_input(request, N < 3)
        db = client.db
        op = request[1].lower()
        if op == b'and':
            reduce_op = and_op
        elif op == b'or':
            reduce_op = or_op
        elif op == b'xor':
            reduce_op = xor_op
        elif op == b'not':
            reduce_op = None
            check_input(request, N != 3)
        else:
            return client.reply_error('bad command')
        empty = bytearray()
        keys = []
        for key in request[3:]:
            value = db.get(key)
            if value is None:
                keys.append(empty)
            elif isinstance(value, bytearray):
                keys.append(value)
            else:
                return client.reply_wrongtype()
        result = bytearray()
        if reduce_op is None:
            for value in keys[0]:
                result.append(~value & 255)
        else:
            for values in zip_longest(*keys, **{'fillvalue': 0}):
                result.append(reduce(reduce_op, values))
        if result:
            dest = request[2]
            if db.pop(dest):
                self._signal(self.NOTIFY_GENERIC, db, 'del', dest)
            db._data[dest] = result
            self._signal(self.NOTIFY_STRING, db, 'set', dest, 1)
            client.reply_int(len(result))
        else:
            client.reply_zero()

    @command('Strings', True)
    def decr(self, client, request, N):
        check_input(request, N != 1)
        r = self._incrby(client, request[0], request[1], b'-1', int)
        client.reply_int(r)

    @command('Strings', True)
    def decrby(self, client, request, N):
        check_input(request, N != 2)
        try:
            val = str(-int(request[2])).encode('utf-8')
        except Exception:
            val = request[2]
        r = self._incrby(client, request[0], request[1], val, int)
        client.reply_int(r)

    @command('Strings')
    def get(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_bulk()
        elif isinstance(value, bytearray):
            assert value
            client.reply_bulk(bytes(value))
        else:
            client.reply_wrongtype()

    @command('Strings')
    def getbit(self, client, request, N):
        check_input(request, N != 2)
        try:
            bitoffset = int(request[2])
            if bitoffset < 0 or bitoffset >= STRING_LIMIT:
                raise ValueError
        except Exception:
            return client.reply_error(
                "bit offset is not an integer or out of range")
        string = client.db.get(request[1])
        if string is None:
            client.reply_zero()
        elif not isinstance(string, bytearray):
            client.reply_wrongtype()
        else:
            assert string
            byte = bitoffset >> 3
            if len(string) > byte:
                bit = 7 - (bitoffset & 7)
                v = string[byte] & (1 << bit)
                client.reply_int(1 if v else 0)
            else:
                client.reply_zero()

    @command('Strings')
    def getrange(self, client, request, N):
        check_input(request, N != 3)
        try:
            start = int(request[2])
            end = int(request[3])
        except Exception:
            return client.reply_error("Wrong offset in '%s' command" %
                                      request[0])
        string = client.db.get(request[1])
        if string is None:
            client.reply_bulk(b'')
        elif not isinstance(string, bytearray):
            client.reply_wrongtype()
        else:
            if start < 0:
                start = len(string) + start
            if end < 0:
                end = len(string) + end + 1
            else:
                end += 1
            client.reply_bulk(bytes(string[start:end]))

    @command('Strings', True)
    def getset(self, client, request, N):
        check_input(request, N != 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            db._data[key] = bytearray(request[2])
            self._signal(self.NOTIFY_STRING, db, 'set', key, 1)
            client.reply_bulk()
        elif isinstance(value, bytearray):
            db.pop(key)
            db._data[key] = bytearray(request[2])
            self._signal(self.NOTIFY_STRING, db, 'set', key, 1)
            client.reply_bulk(bytes(value))
        else:
            client.reply_wrongtype()

    @command('Strings', True)
    def incr(self, client, request, N):
        check_input(request, N != 1)
        r = self._incrby(client, request[0], request[1], b'1', int)
        client.reply_int(r)

    @command('Strings', True)
    def incrby(self, client, request, N):
        check_input(request, N != 2)
        r = self._incrby(client, request[0], request[1], request[2], int)
        client.reply_int(r)

    @command('Strings', True)
    def incrbyfloat(self, client, request, N):
        check_input(request, N != 2)
        r = self._incrby(client, request[0], request[1], request[2], float)
        client.reply_bulk(str(r).encode('utf-8'))

    @command('Strings')
    def mget(self, client, request, N):
        check_input(request, not N)
        get = client.db.get
        values = []
        for key in request[1:]:
            value = get(key)
            if value is None:
                values.append(value)
            elif isinstance(value, bytearray):
                values.append(bytes(value))
            else:
                return client.reply_wrongtype()
        client.reply_multi_bulk(values)

    @command('Strings', True)
    def mset(self, client, request, N):
        D = N // 2
        check_input(request, N < 2 or D * 2 != N)
        db = client.db
        for key, value in zip(request[1::2], request[2::2]):
            db.pop(key)
            db._data[key] = bytearray(value)
            self._signal(self.NOTIFY_STRING, db, 'set', key, 1)
        client.reply_ok()

    @command('Strings', True)
    def msetnx(self, client, request, N):
        D = N // 2
        check_input(request, N < 2 or D * 2 != N)
        db = client.db
        keys = request[1::2]
        exist = False
        for key in keys:
            exist = db.exists(key)
            if exist:
                break
        if exist:
            client.reply_zero()
        else:
            for key, value in zip(keys, request[2::2]):
                db._data[key] = bytearray(value)
                self._signal(self.NOTIFY_STRING, db, 'set', key, 1)
            client.reply_one()

    @command('Strings', True)
    def psetex(self, client, request, N):
        check_input(request, N != 3)
        self._set(client, request[1], request[3], milliseconds=request[2])
        client.reply_ok()

    @command('Strings', True)
    def set(self, client, request, N):
        check_input(request, N < 2 or N > 8)
        it = 2
        extra = set(self._set_options)
        seconds = 0
        milliseconds = 0
        nx = False
        xx = False
        while N > it:
            it += 1
            opt = request[it].lower()
            if opt in extra:
                extra.remove(opt)
                if opt == b'ex':
                    it += 1
                    seconds = request[it]
                elif opt == b'px':
                    it += 1
                    milliseconds = request[it]
                elif opt == b'nx':
                    nx = True
                else:
                    xx = True
        if self._set(client, request[1], request[2], seconds,
                     milliseconds, nx, xx):
            client.reply_ok()
        else:
            client.reply_bulk()

    @command('Strings', True)
    def setbit(self, client, request, N):
        check_input(request, N != 3)
        key = request[1]
        try:
            bitoffset = int(request[2])
            if bitoffset < 0 or bitoffset >= STRING_LIMIT:
                raise ValueError
        except Exception:
            return client.reply_error(
                "bit offset is not an integer or out of range")
        try:
            value = int(request[3])
            if value not in (0, 1):
                raise ValueError
        except Exception:
            return client.reply_error("bit is not an integer or out of range")
        db = client.db
        string = db.get(key)
        if string is None:
            string = bytearray()
            db._data[key] = string
        elif not isinstance(string, bytearray):
            return client.reply_wrongtype()
        else:
            assert string

        # grow value to the right if necessary
        byte = bitoffset >> 3
        num_bytes = len(string)
        if byte >= num_bytes:
            string.extend((byte + 1 - num_bytes)*b'\x00')

        # get current value
        byteval = string[byte]
        bit = 7 - (bitoffset & 7)
        bitval = byteval & (1 << bit)

        # update with new value
        byteval &= ~(1 << bit)
        byteval |= ((value & 1) << bit)
        string[byte] = byteval

        self._signal(self.NOTIFY_STRING, db, request[0], key, 1)
        client.reply_one() if bitval else client.reply_zero()

    @command('Strings', True)
    def setex(self, client, request, N):
        check_input(request, N != 3)
        self._set(client, request[1], request[3], seconds=request[2])
        client.reply_ok()

    @command('Strings', True)
    def setnx(self, client, request, N):
        check_input(request, N != 2)
        if self._set(client, request[1], request[2], nx=True):
            client.reply_one()
        else:
            client.reply_zero()

    @command('Strings', True)
    def setrange(self, client, request, N):
        check_input(request, N != 3)
        key = request[1]
        value = request[3]
        try:
            offset = int(request[2])
            T = offset + len(value)
            if offset < 0 or T >= STRING_LIMIT:
                raise ValueError
        except Exception:
            return client.reply_error("Wrong offset in '%s' command" %
                                      request[0])
        db = client.db
        string = db.get(key)
        if string is None:
            string = bytearray(b'\x00')
            db._data[key] = string
        elif not isinstance(string, bytearray):
            return client.reply_wrongtype()
        N = len(string)
        if N < T:
            string.extend((T + 1 - N)*b'\x00')
        string[offset:T] = value
        self._signal(self.NOTIFY_STRING, db, request[0], key, 1)
        client.reply_int(len(string))

    @command('Strings')
    def strlen(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif isinstance(value, bytearray):
            client.reply_int(len(value))
        else:
            return client.reply_wrongtype()

    # #########################################################################
    # #    HASHES COMMANDS
    @command('Hashes', True)
    def hdel(self, client, request, N):
        check_input(request, N < 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif isinstance(value, self.hash_type):
            rem = 0
            for field in request[2:]:
                rem += 0 if value.pop(field, None) is None else 1
            self._signal(self.NOTIFY_HASH, db, request[0], key, rem)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_int(rem)
        else:
            client.reply_wrongtype()

    @command('Hashes')
    def hexists(self, client, request, N):
        check_input(request, N != 2)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif isinstance(value, self.hash_type):
            client.reply_int(int(request[2] in value))
        else:
            client.reply_wrongtype()

    @command('Hashes')
    def hget(self, client, request, N):
        check_input(request, N != 2)
        value = client.db.get(request[1])
        if value is None:
            client.reply_bulk()
        elif isinstance(value, self.hash_type):
            client.reply_bulk(value.get(request[2]))
        else:
            client.reply_wrongtype()

    @command('Hashes')
    def hgetall(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif isinstance(value, self.hash_type):
            client.reply_multi_bulk(value.flat())
        else:
            client.reply_wrongtype()

    @command('Hashes', True)
    def hincrby(self, client, request, N):
        result = self._hincrby(client, request, N, int)
        if result is not None:
            client.reply_int(result)

    @command('Hashes', True)
    def hincrbyfloat(self, client, request, N):
        result = self._hincrby(client, request, N, float)
        if result is not None:
            client.reply_bulk(str(result).encode('utf-8'))

    @command('Hashes')
    def hkeys(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif isinstance(value, self.hash_type):
            client.reply_multi_bulk(value)
        else:
            client.reply_wrongtype()

    @command('Hashes')
    def hlen(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif isinstance(value, self.hash_type):
            client.reply_int(len(value))
        else:
            client.reply_wrongtype()

    @command('Hashes')
    def hmget(self, client, request, N):
        check_input(request, N < 3)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif isinstance(value, self.hash_type):
            result = value.mget(request[2:])
            client.reply_multi_bulk(result)
        else:
            client.reply_wrongtype()

    @command('Hashes', True)
    def hmset(self, client, request, N):
        D = (N - 1) // 2
        check_input(request, N < 3 or D * 2 != N - 1)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            value = self.hash_type()
            db._data[key] = value
        elif not isinstance(value, self.hash_type):
            return client.reply_wrongtype()
        it = iter(request[2:])
        value.update(zip(it, it))
        self._signal(self.NOTIFY_HASH, db, request[0], key, D)
        client.reply_ok()

    @command('Hashes', True)
    def hset(self, client, request, N):
        check_input(request, N != 3)
        key, field = request[1], request[2]
        db = client.db
        value = db.get(key)
        if value is None:
            value = self.hash_type()
            db._data[key] = value
        elif not isinstance(value, self.hash_type):
            return client.reply_wrongtype()
        avail = (field in value)
        value[field] = request[3]
        self._signal(self.NOTIFY_HASH, db, request[0], key, 1)
        client.reply_zero() if avail else client.reply_one()

    @command('Hashes', True)
    def hsetnx(self, client, request, N):
        check_input(request, N != 3)
        key, field = request[1], request[2]
        db = client.db
        value = db.get(key)
        if value is None:
            value = self.hash_type()
            db._data[key] = value
        elif not isinstance(value, self.hash_type):
            return client.reply_wrongtype()
        if field in value:
            client.reply_zero()
        else:
            value[field] = request[3]
            self._signal(self.NOTIFY_HASH, db, request[0], key, 1)
            client.reply_one()

    @command('Hashes')
    def hvals(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif isinstance(value, self.hash_type):
            client.reply_multi_bulk(tuple(value.values()))
        else:
            client.reply_wrongtype()

    @command('Hashes', supported=False)
    def hscan(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    # #########################################################################
    # #    LIST COMMANDS
    @command('Lists', True, script=0)
    def blpop(self, client, request, N):
        check_input(request, N < 2)
        try:
            timeout = max(0, int(request[-1]))
        except Exception:
            return client.reply_error(self.SYNTAX_ERROR)
        keys = request[1:-1]
        if not self._bpop(client, request, keys):
            client.blocked = Blocked(client, request[0], keys, timeout)

    @command('Lists', True, script=0)
    def brpop(self, client, request, N):
        return self.blpop(client, request, N)

    @command('Lists', True, script=0)
    def brpoplpush(self, client, request, N):
        check_input(request, N != 3)
        try:
            timeout = max(0, int(request[-1]))
        except Exception:
            return client.reply_error(self.SYNTAX_ERROR)
        key, dest = request[1:-1]
        keys = (key,)
        if not self._bpop(client, request, keys, dest):
            client.blocked = Blocked(client, request[0], keys, timeout, dest)

    @command('Lists')
    def lindex(self, client, request, N):
        check_input(request, N != 2)
        value = client.db.get(request[1])
        if value is None:
            client.reply_bulk()
        elif isinstance(value, self.list_type):
            assert value
            index = int(request[2])
            if index >= 0 and index < len(value):
                client.reply_bulk(value[index])
            else:
                client.reply_bulk()
        else:
            client.reply_wrongtype()

    @command('Lists', True)
    def linsert(self, client, request, N):
        # This method is INEFFICIENT, but redis supported so we do
        # the same here
        check_input(request, N != 4)
        db = client.db
        key = request[1]
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            where = request[2].lower()
            l1 = len(value)
            if where == b'before':
                value.insert_before(request[3], request[4])
            elif where == b'after':
                value.insert_after(request[3], request[4])
            else:
                return client.reply_error('cannot insert to list')
            l2 = len(value)
            if l2 - l1:
                self._signal(self.NOTIFY_LIST, db, request[0], key, 1)
                client.reply_int(l2)
            else:
                client.reply_int(-1)

    @command('Lists')
    def llen(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif isinstance(value, self.list_type):
            assert value
            client.reply_int(len(value))
        else:
            client.reply_wrongtype()

    @command('Lists', True)
    def lpop(self, client, request, N):
        check_input(request, N != 1)
        db = client.db
        key = request[1]
        value = db.get(key)
        if value is None:
            client.reply_bulk()
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            if request[0] == 'lpop':
                result = value.popleft()
            else:
                result = value.pop()
            self._signal(self.NOTIFY_LIST, db, request[0], key, 1)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_bulk(result)

    @command('Lists', True)
    def rpop(self, client, request, N):
        return self.lpop(client, request, N)

    @command('Lists', True)
    def lpush(self, client, request, N):
        check_input(request, N < 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            value = self.list_type()
            db._data[key] = value
        elif not isinstance(value, self.list_type):
            return client.reply_wrongtype()
        else:
            assert value
        if request[0] == 'lpush':
            value.extendleft(request[2:])
        else:
            value.extend(request[2:])
        client.reply_int(len(value))
        self._signal(self.NOTIFY_LIST, db, request[0], key, N - 1)

    @command('Lists', True)
    def rpush(self, client, request, N):
        return self.lpush(client, request, N)

    @command('Lists', True)
    def lpushx(self, client, request, N):
        check_input(request, N != 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            if request[0] == 'lpushx':
                value.appendleft(request[2])
            else:
                value.append(request[2])
            client.reply_int(len(value))
            self._signal(self.NOTIFY_LIST, db, request[0], key, 1)

    @command('Lists', True)
    def rpushx(self, client, request, N):
        return self.lpushx(client, request, N)

    @command('Lists', True)
    def lrange(self, client, request, N):
        check_input(request, N != 3)
        db = client.db
        key = request[1]
        value = db.get(key)
        try:
            start, end = self._range_values(value, request[2], request[3])
        except Exception:
            return client.reply_error('invalid range')
        if value is None:
            client.reply_multi_bulk(())
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            client.reply_multi_bulk(tuple(islice(value, start, end)))

    @command('Lists', True)
    def lrem(self, client, request, N):
        # This method is INEFFICIENT, but redis supported so we do
        # the same here
        check_input(request, N != 3)
        db = client.db
        key = request[1]
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            try:
                count = int(request[2])
            except Exception:
                return client.reply_error('cannot remove from list')
            removed = value.remove(request[3], count)
            if removed:
                self._signal(self.NOTIFY_LIST, db, request[0], key, removed)
            client.reply_int(removed)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)

    @command('Lists', True)
    def lset(self, client, request, N):
        check_input(request, N != 3)
        db = client.db
        key = request[1]
        value = db.get(key)
        if value is None:
            client.reply_error(self.OUT_OF_BOUND)
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            try:
                index = int(request[2])
            except Exception:
                index = -1
            if index >= 0 and index < len(value):
                value[index] = request[3]
                self._signal(self.NOTIFY_LIST, db, request[0], key, 1)
                client.reply_ok()
            else:
                client.reply_error(self.OUT_OF_BOUND)

    @command('Lists', True)
    def ltrim(self, client, request, N):
        check_input(request, N != 3)
        db = client.db
        key = request[1]
        value = db.get(key)
        try:
            start, end = self._range_values(value, request[2], request[3])
        except Exception:
            return client.reply_error('invalid range')
        if value is None:
            client.reply_ok()
        elif not isinstance(value, self.list_type):
            client.reply_wrongtype()
        else:
            assert value
            start = len(value)
            value.trim(start, end)
            client.reply_ok()
            self._signal(self.NOTIFY_LIST, db, request[0], key,
                         start-len(value))
            client.reply_ok()
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)

    @command('Lists', True)
    def rpoplpush(self, client, request, N):
        check_input(request, N != 2)
        key1, key2 = request[1], request[2]
        db = client.db
        orig = db.get(key1)
        dest = db.get(key2)
        if orig is None:
            client.reply_bulk()
        elif not isinstance(orig, self.list_type):
            client.reply_wrongtype()
        else:
            assert orig
            if dest is None:
                dest = self.list_type()
                db._data[key2] = dest
            elif not isinstance(dest, self.list_type):
                return client.reply_wrongtype()
            else:
                assert dest
            value = orig.pop()
            self._signal(self.NOTIFY_LIST, db, 'rpop', key1, 1)
            dest.appendleft(value)
            self._signal(self.NOTIFY_LIST, db, 'lpush', key2, 1)
            if db.pop(key1, orig) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key1)
            client.reply_bulk(value)

    # #########################################################################
    # #    SETS COMMANDS
    @command('Sets', True)
    def sadd(self, client, request, N):
        check_input(request, N < 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            value = set()
            db._data[key] = value
        elif not isinstance(value, set):
            return client.reply_wrongtype()
        n = len(value)
        value.update(request[2:])
        n = len(value) - n
        self._signal(self.NOTIFY_SET, db, request[0], key, n)
        client.reply_int(n)

    @command('Sets')
    def scard(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif not isinstance(value, set):
            client.reply_wrongtype()
        else:
            client.reply_int(len(value))

    @command('Sets')
    def sdiff(self, client, request, N):
        check_input(request, N < 1)
        self._setoper(client, 'difference', request[1:])

    @command('Sets', True)
    def sdiffstore(self, client, request, N):
        check_input(request, N < 2)
        self._setoper(client, 'difference', request[2:], request[1])

    @command('Sets')
    def sinter(self, client, request, N):
        check_input(request, N < 1)
        self._setoper(client, 'intersection', request[1:])

    @command('Sets', True)
    def sinterstore(self, client, request, N):
        check_input(request, N < 2)
        self._setoper(client, 'intersection', request[2:], request[1])

    @command('Sets')
    def sismember(self, client, request, N):
        check_input(request, N != 2)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif not isinstance(value, set):
            client.reply_wrongtype()
        else:
            client.reply_int(int(request[2] in value))

    @command('Sets')
    def smembers(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif not isinstance(value, set):
            client.reply_wrongtype()
        else:
            client.reply_multi_bulk(value)

    @command('Sets', True)
    def smove(self, client, request, N):
        check_input(request, N != 3)
        db = client.db
        key1 = request[1]
        key2 = request[2]
        orig = db.get(key1)
        dest = db.get(key2)
        if orig is None:
            client.reply_zero()
        elif not isinstance(orig, set):
            client.reply_wrongtype()
        else:
            member = request[3]
            if member in orig:
                # we my be able to move
                if dest is None:
                    dest = set()
                    db._data[request[2]] = dest
                elif not isinstance(dest, set):
                    return client.reply_wrongtype()
                orig.remove(member)
                dest.add(member)
                self._signal(self.NOTIFY_SET, db, 'srem', key1)
                self._signal(self.NOTIFY_SET, db, 'sadd', key2, 1)
                if db.pop(key1, orig) is not None:
                    self._signal(self.NOTIFY_GENERIC, db, 'del', key1)
                client.reply_one()
            else:
                client.reply_zero()

    @command('Sets', True)
    def spop(self, client, request, N):
        check_input(request, N != 1)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_bulk()
        elif not isinstance(value, set):
            client.reply_wrongtype()
        else:
            result = value.pop()
            self._signal(self.NOTIFY_SET, db, request[0], key, 1)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_bulk(result)

    @command('Sets')
    def srandmember(self, client, request, N):
        check_input(request, N < 1 or N > 2)
        value = client.db.get(request[1])
        if value is not None and not isinstance(value, set):
            return client.reply_wrongtype()
        if N == 2:
            try:
                count = int(request[2])
            except Exception:
                return client.reply_error('Invalid count')
            if count < 0:
                count = -count
                if not value:
                    result = (None,) * count
                else:
                    result = []
                    for _ in range(count):
                        el = value.pop()
                        result.append(el)
                        value.add(el)
            elif count > 0:
                if not value:
                    result = (None,)
                elif len(value) <= count:
                    result = list(value)
                    result.extend((None,)*(count-len(value)))
                else:
                    result = []
                    for _ in range(count):
                        el = value.pop()
                        result.append(el)
                    value.update(result)
            else:
                result = []
            client.reply_multi_bulk(result)
        else:
            if not value:
                result = None
            else:
                result = value.pop()
                value.add(result)
            client.reply_bulk(result)

    @command('Sets', True)
    def srem(self, client, request, N):
        check_input(request, N < 2)
        db = client.db
        key = request[1]
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, set):
            client.reply_wrongtype()
        else:
            start = len(value)
            value.difference_update(request[2:])
            removed = start - len(value)
            self._signal(self.NOTIFY_SET, db, request[0], key, removed)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_int(removed)

    @command('Sets')
    def sunion(self, client, request, N):
        check_input(request, N < 1)
        self._setoper(client, 'union', request[1:])

    @command('Sets', True)
    def sunionstore(self, client, request, N):
        check_input(request, N < 2)
        self._setoper(client, 'union', request[2:], request[1])

    @command('Sets', supported=False)
    def sscan(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    # #########################################################################
    # #    SORTED SETS COMMANDS
    @command('Sorted Sets', True)
    def zadd(self, client, request, N):
        D = (N - 1) // 2
        check_input(request, N < 3 or D * 2 != N - 1)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            value = self.zset_type()
            db._data[key] = value
        elif not isinstance(value, self.zset_type):
            return client.reply_wrongtype()
        start = len(value)
        value.update(zip(map(float, request[2::2]), request[3::2]))
        result = len(value) - start
        self._signal(self.NOTIFY_ZSET, db, request[0], key, result)
        client.reply_int(result)

    @command('Sorted Sets')
    def zcard(self, client, request, N):
        check_input(request, N != 1)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            client.reply_int(len(value))

    @command('Sorted Sets')
    def zcount(self, client, request, N):
        check_input(request, N != 3)
        value = client.db.get(request[1])
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            min_value, max_value = request[2], request[3]
            include_min = include_max = True
            if min_value and _ord(min_value[0]) == 40:
                include_min = False
                min_value = min_value[1:]
            if max_value and _ord(max_value[0]) == 40:
                include_max = False
                max_value = max_value[1:]
            try:
                mmin = float(min_value)
                mmax = float(max_value)
            except Exception:
                client.reply_error(self.INVALID_SCORE)
            else:
                client.reply_int(value.count(mmin, mmax,
                                             include_min, include_max))

    @command('Sorted Sets', True)
    def zincrby(self, client, request, N):
        check_input(request, N != 3)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            db._data[key] = value = self.zset_type()
        elif not isinstance(value, self.zset_type):
            return client.reply_wrongtype()
        try:
            increment = float(request[2])
        except Exception:
            client.reply_error(self.INVALID_SCORE)
        else:
            member = request[3]
            score = value.score(member, 0) + increment
            value.add(score, member)
            self._signal(self.NOTIFY_ZSET, db, request[0], key, 1)
            client.reply_bulk(str(score).encode('utf-8'))

    @command('Sorted Sets', True)
    def zinterstore(self, client, request, N):
        self._zsetoper(client, request, N)

    @command('Sorted Sets')
    def zrange(self, client, request, N):
        check_input(request, N < 3 or N > 4)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            try:
                start, end = self._range_values(value, request[2], request[3])
            except Exception:
                return client.reply_error(self.SYNTAX_ERROR)
            reverse = (request[0] == b'zrevrange')
            if N == 4:
                if request[4].lower() == b'withscores':
                    result = []
                    [result.extend((v, score)) for score, v in
                     value.range(start, end, scores=True)]
                else:
                    return client.reply_error(self.SYNTAX_ERROR)
            else:
                result = list(value.range(start, end))
            client.reply_multi_bulk(result)

    @command('Sorted Sets')
    def zrangebyscore(self, client, request, N):
        check_input(request, N < 3 or N > 7)
        value = client.db.get(request[1])
        if value is None:
            client.reply_multi_bulk(())
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            try:
                minval, include_min, maxval, include_max = self._score_values(
                    request[2], request[3])
            except Exception:
                return client.reply_error(self.SYNTAX_ERROR)
            request = request[4:]
            withscores = False
            offset = 0
            count = None
            while request:
                if request[0].lower() == b'withscores':
                    withscores = True
                    request = request[1:]
                elif request[0].lower() == b'limit':
                    try:
                        offset = int(request[1])
                        count = int(request[2])
                    except Exception:
                        return client.reply_error(self.SYNTAX_ERROR)
                    request = request[3:]
                else:
                    return client.reply_error(self.SYNTAX_ERROR)
            if withscores:
                result = []
                [result.extend((v, score)) for score, v in
                 value.range_by_score(minval, maxval, scores=True,
                                      start=offset, num=count,
                                      include_min=include_min,
                                      include_max=include_max)]
            else:
                result = list(value.range_by_score(minval, maxval,
                                                   start=offset, num=count,
                                                   include_min=include_min,
                                                   include_max=include_max))
            client.reply_multi_bulk(result)

    @command('Sorted Sets')
    def zrank(self, client, request, N):
        check_input(request, N != 2)
        value = client.db.get(request[1])
        if value is None:
            client.reply_bulk()
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            rank = value.rank(request[2])
            if rank is not None:
                client.reply_int(rank)
            else:
                client.reply_bulk()

    @command('Sorted Sets', True)
    def zrem(self, client, request, N):
        check_input(request, N < 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            removed = value.remove_items(request[2:])
            if removed:
                self._signal(self.NOTIFY_ZSET, db, request[0], key, removed)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_int(removed)

    @command('Sorted Sets', True)
    def zremrangebyrank(self, client, request, N):
        check_input(request, N != 3)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            try:
                start, end = self._range_values(value, request[2], request[3])
            except Exception:
                return client.reply_error(self.SYNTAX_ERROR)
            removed = value.remove_range(start, end)
            if removed:
                self._signal(self.NOTIFY_ZSET, db, request[0], key, removed)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_int(removed)

    @command('Sorted Sets', True)
    def zremrangebyscore(self, client, request, N):
        check_input(request, N != 3)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_zero()
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            try:
                minval, include_min, maxval, include_max = self._score_values(
                    request[2], request[3])
            except Exception:
                return client.reply_error(self.SYNTAX_ERROR)
            removed = value.remove_range_by_score(minval, maxval, include_min,
                                                  include_max)
            if removed:
                self._signal(self.NOTIFY_ZSET, db, request[0], key, removed)
            if db.pop(key, value) is not None:
                self._signal(self.NOTIFY_GENERIC, db, 'del', key)
            client.reply_int(removed)

    @command('Sorted Sets', supported=False)
    def zrevrange(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Sorted Sets', supported=False)
    def zrevrangebyscore(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Sorted Sets')
    def zscore(self, client, request, N):
        check_input(request, N != 2)
        key = request[1]
        db = client.db
        value = db.get(key)
        if value is None:
            client.reply_bulk(None)
        elif not isinstance(value, self.zset_type):
            client.reply_wrongtype()
        else:
            score = value.score(request[2], None)
            if score is not None:
                score = str(score).encode('utf-8')
            client.reply_bulk(score)

    @command('Sorted Sets', True)
    def zunionstore(self, client, request, N):
        self._zsetoper(client, request, N)

    @command('Sorted Sets', supported=False)
    def zscan(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    # #########################################################################
    # #    PUBSUB COMMANDS
    @command('Pub/Sub', script=0)
    def psubscribe(self, client, request, N):
        check_input(request, not N)
        for pattern in request[1:]:
            p = self._patterns.get(pattern)
            if not p:
                pre = redis_to_py_pattern(pattern.decode('utf-8'))
                p = pubsub_patterns(re.compile(pre), set())
                self._patterns[pattern] = p
            p.clients.add(client)
            client.patterns.add(pattern)
            count = reduce(lambda x, y: x + int(client in y.clients),
                           self._patterns.values())
            client.reply_multi_bulk((b'psubscribe', pattern, count))

    @command('Pub/Sub')
    def pubsub(self, client, request, N):
        check_input(request, not N)
        subcommand = request[1].decode('utf-8').lower()
        if subcommand == 'channels':
            check_input(request, N > 2)
            if N == 2:
                pre = re.compile(redis_to_py_pattern(
                    request[2].decode('utf-8')))
                channels = []
                for channel in self._channels:
                    if pre.match(channel.decode('utf-8', 'ignore')):
                        channels.append(channel)
            else:
                channels = list(self._channels)
            client.reply_multi_bulk(channels)
        elif subcommand == 'numsub':
            count = []
            for channel in request[2:]:
                clients = self._channels.get(channel, ())
                count.extend((channel, len(clients)))
            client.reply_multi_bulk(count)
        elif subcommand == 'numpat':
            check_input(request, N > 1)
            count = reduce(lambda x, y: x + len(y.clients),
                           self._patterns.values())
            client.reply_int(count)
        else:
            client.reply_error("Unknown command 'pubsub %s'" % subcommand)

    @command('Pub/Sub')
    def publish(self, client, request, N):
        check_input(request, N != 2)
        channel, message = request[1:]
        ch = channel.decode('utf-8')
        msg = self._parser.multi_bulk((b'message', channel, message))
        count = self._publish_clients(msg, self._channels.get(channel, ()))
        for pattern in self._patterns.values():
            g = pattern.re.match(ch)
            if g:
                count += self._publish_clients(msg, pattern.clients)
        client.reply_int(count)

    @command('Pub/Sub', script=0)
    def punsubscribe(self, client, request, N):
        patterns = request[1:] if N else list(self._patterns)
        for pattern in patterns:
            if pattern in self._patterns:
                p = self._patterns[pattern]
                if client in p.clients:
                    client.patterns.discard(pattern)
                    p.clients.remove(client)
                    if not p.clients:
                        self._patterns.pop(pattern)
                    client.reply_multi_bulk((b'punsubscribe', pattern))

    @command('Pub/Sub', script=0)
    def subscribe(self, client, request, N):
        check_input(request, not N)
        for channel in request[1:]:
            clients = self._channels.get(channel)
            if not clients:
                self._channels[channel] = clients = set()
            clients.add(client)
            client.channels.add(channel)
            client.reply_multi_bulk((b'subscribe', channel, len(clients)))

    @command('Pub/Sub', script=0)
    def unsubscribe(self, client, request, N):
        channels = request[1:] if N else list(self._channels)
        for channel in channels:
            if channel in self._channels:
                clients = self._channels[channel]
                if client in clients:
                    client.channels.discard(channel)
                    clients.remove(client)
                    if not clients:
                        self._channels.pop(channel)
                    client.reply_multi_bulk((b'unsubscribe', channel))

    # #########################################################################
    # #    TRANSACTION COMMANDS
    @command('Transactions', script=0)
    def discard(self, client, request, N):
        check_input(request, N)
        if client.transaction is None:
            client.reply_error("DISCARD without MULTI")
        else:
            self._close_transaction(client)
            client.reply_ok()

    @command('Transactions', name='exec', script=0)
    def execute(self, client, request, N):
        check_input(request, N)
        if client.transaction is None:
            client.reply_error("EXEC without MULTI")
        else:
            requests = client.transaction
            if client.flag & self.DIRTY_CAS:
                self._close_transaction(client)
                client.reply_multi_bulk(())
            else:
                self._close_transaction(client)
                client.reply_multi_bulk_len(len(requests))
                for handle, request in requests:
                    client._execute_command(handle, request)

    @command('Transactions', script=0)
    def multi(self, client, request, N):
        check_input(request, N)
        if client.transaction is None:
            client.reply_ok()
            client.transaction = []
        else:
            self.error_replay("MULTI calls can not be nested")

    @command('Transactions', script=0)
    def watch(self, client, request, N):
        check_input(request, not N)
        if client.transaction is not None:
            client.reply_error("WATCH inside MULTI is not allowed")
        else:
            wkeys = client.watched_keys
            if not wkeys:
                client.watched_keys = wkeys = set()
                self._watching.add(client)
            wkeys.update(request[1:])
            client.reply_ok()

    @command('Transactions', script=0)
    def unwatch(self, client, request, N):
        check_input(request, N)
        transaction = client.transaction
        self._close_transaction(client)
        client.transaction = transaction
        client.reply_ok()

    # #########################################################################
    # #    SCRIPTING
    @command('Scripting', script=0)
    def eval(self, client, request, N):
        check_input(request, N < 2)
        if not self.lua:
            return client.reply_error(self.SYNTAX_ERROR)
        script = request[1]
        self._eval_script(client, script, request)

    @command('Scripting', script=0)
    def evalsha(self, client, request, N):
        check_input(request, N < 2)
        if not self.lua:
            return client.reply_error(self.SYNTAX_ERROR)
        script = self.scripts.get(request[2])
        if script is None:
            client.reply_error('the script is not available', 'NOSCRIPT')
        else:
            self._eval_script(client, script, request)

    @command('Scripting', script=0)
    def script(self, client, request, N):
        check_input(request, not N)
        if not self.lua:
            return client.reply_error(self.SYNTAX_ERROR)
        scripts = self.scripts
        subcommand = request[1].upper()
        if subcommand == b'EXISTS':
            check_input(request, N < 2)
            result = [int(sha in scripts) for sha in request[2:]]
            client.reply_multi_bulk(result)
        elif subcommand == b'FLUSH':
            check_input(request, N != 1)
            scripts.clear()
            client.reply_ok()
        elif subcommand == b'KILL':
            check_input(request, N != 1)
            client.reply_ok()
        elif subcommand == b'LOAD':
            check_input(request, N != 2)
            script = request[2]
            sha = sha1(script).hexdigest().encode('utf-8')
            scripts[sha] = script
            client.reply_bulk(sha)

    # #########################################################################
    # #    CONNECTION COMMANDS
    @command('Connections', script=0)
    def auth(self, client, request, N):
        check_input(request, N != 1)
        client.password = request[1]
        if client.password != client._producer.password:
            client.reply_error("wrong password")
        else:
            client.reply_ok()

    @command('Connections')
    def echo(self, client, request, N):
        check_input(request, N != 1)
        client.reply_bulk(request[1])

    @command('Connections')
    def ping(self, client, request, N):
        check_input(request, N)
        client.reply_status('PONG')

    @command('Connections', script=0)
    def quit(self, client, request, N):
        check_input(request, N)
        client.reply_ok()
        client.close()

    @command('Connections')
    def select(self, client, request, N):
        check_input(request, N != 1)
        D = len(self.databases) - 1
        try:
            num = int(request[1])
            if num < 0 or num > D:
                raise ValueError
        except ValueError:
            client.reply_error(('select requires a database number between '
                                '%s and %s' % (0, D)))
        else:
            client.database = num
            client.reply_ok()

    # #########################################################################
    # #    SERVER COMMANDS
    @command('Server', supported=False)
    def bgrewriteaof(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Server')
    def bgsave(self, client, request, N):
        check_input(request, N)
        self._save()
        client.reply_ok()

    @command('Server')
    def client(self, client, request, N):
        check_input(request, not N)
        subcommand = request[1].decode('utf-8').lower()
        if subcommand == 'list':
            check_input(request, N != 1)
            value = '\n'.join(self._client_list(client))
            client.reply_bulk(value.encode('utf-8'))
        else:
            client.reply_error("unknown command 'client %s'" % subcommand)

    @command('Server')
    def config(self, client, request, N):
        check_input(request, not N)
        subcommand = request[1].decode('utf-8').lower()
        if subcommand == 'get':
            if N != 2:
                client.reply_error("'config get' no argument")
            else:
                value = self._get_config(request[2].decode('utf-8'))
                client.reply_bulk(value)
        elif subcommand == 'rewrite':
            client.reply_ok()
        elif subcommand == 'set':
            try:
                if N != 3:
                    raise ValueError("'config set' no argument")
                self._set_config(request[2].decode('utf-8'))
            except Exception as e:
                client.reply_error(str(e))
            else:
                client.reply_ok()
        elif subcommand == 'resetstat':
            self._hit_keys = 0
            self._missed_keys = 0
            self._expired_keys = 0
            server = client._producer
            server._received = 0
            server._requests_processed = 0
            client.reply_ok()
        else:
            client.reply_error("'config %s' not valid" % subcommand)

    @command('Server')
    def dbsize(self, client, request, N):
        check_input(request, N != 0)
        client.reply_int(len(client.db))

    @command('Server', supported=False, subcommands=['object', 'segfault'])
    def debug(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Server', True)
    def flushdb(self, client, request, N):
        check_input(request, N)
        client.db.flush()
        client.reply_ok()

    @command('Server', True)
    def flushall(self, client, request, N):
        check_input(request, N)
        for db in self.databases.values():
            db.flush()
        client.reply_ok()

    @command('Server')
    def info(self, client, request, N):
        check_input(request, N)
        info = '\n'.join(self._flat_info())
        client.reply_bulk(info.encode('utf-8'))

    @command('Server')
    def lastsave(self, client, request, N):
        check_input(request, N)
        client.reply_int(self._last_save)

    @command('Server', script=0)
    def monitor(self, client, request, N):
        check_input(request, N)
        client.flag |= self.MONITOR
        self._monitors.add(client)
        client.reply_ok()

    @command('Server', script=0)
    def save(self, client, request, N):
        check_input(request, N)
        self._save(False)
        client.reply_ok()

    @command('Server', supported=False)
    def shutdown(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Server', supported=False)
    def slaveof(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Server', supported=False)
    def slowlog(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Server', supported=False)
    def sync(self, client, request, N):
        client.reply_error(self.NOT_SUPPORTED)

    @command('Server')
    def time(self, client, request, N):
        check_input(request, N != 0)
        t = time.time()
        seconds = math.floor(t)
        microseconds = int(1000000*(t-seconds))
        client.reply_multi_bulk((seconds, microseconds))

    # #########################################################################
    # #    INTERNALS
    def _cron(self):
        dirty = self._dirty
        if dirty:
            now = time.time()
            gap = now - self._last_save
            for interval, changes in self.cfg.key_value_save:
                if gap >= interval and dirty >= changes:
                    self._save()
                    break

    def _set(self, client, key, value, seconds=0, milliseconds=0,
             nx=False, xx=False):
        try:
            seconds = int(seconds)
            milliseconds = 0.000001*int(milliseconds)
            if seconds < 0 or milliseconds < 0:
                raise ValueError
        except Exception:
            return client.reply_error('invalid expire time')
        else:
            timeout = seconds + milliseconds
        db = client.db
        exists = db.exists(key)
        skip = (exists and nx) or (not exists and xx)
        if not skip:
            if exists:
                db.pop(key)
            if timeout > 0:
                db._expires[key] = (self._loop.call_later(
                    timeout, db._do_expire, key), bytearray(value))
                self._signal(self.NOTIFY_STRING, db, 'expire', key)
            else:
                db._data[key] = bytearray(value)
            self._signal(self.NOTIFY_STRING, db, 'set', key, 1)
            return True

    def _incrby(self, client, name, key, value, type):
        try:
            tv = type(value)
        except Exception:
            return client.reply_error('invalid increment')
        db = client.db
        cur = db.get(key)
        if cur is None:
            db._data[key] = bytearray(value)
        elif isinstance(cur, bytearray):
            try:
                tv += type(cur)
            except Exception:
                return client.reply_error('invalid increment')
            db._data[key] = bytearray(str(tv).encode('utf-8'))
        else:
            return client.reply_wrongtype()
        self._signal(self.NOTIFY_STRING, db, name, key, 1)
        return tv

    def _bpop(self, client, request, keys, dest=None):
        list_type = self.list_type
        db = client.db
        for key in keys:
            value = db.get(key)
            if isinstance(value, list_type):
                self._block_callback(client, request[0], key, value, dest)
                return True
            elif value is not None:
                client.reply_wrongtype()
                return True
        return False

    def _block_callback(self, client, command, key, value, dest):
        db = client.db
        if command[:2] == 'br':
            if dest is not None:
                dval = db.get(dest)
                if dval is None:
                    dval = self.list_type()
                    db._data[dest] = dval
                elif not isinstance(dval, self.list_type):
                    return client.reply_wrongtype()
            elem = value.pop()
            self._signal(self.NOTIFY_LIST, db, 'rpop', key, 1)
            if dest is not None:
                dval.appendleft(elem)
                self._signal(self.NOTIFY_LIST, db, 'lpush', dest, 1)
        else:
            elem = value.popleft()
            self._signal(self.NOTIFY_LIST, db, 'lpop', key, 1)
        if not value:
            db.pop(key)
            self._signal(self.NOTIFY_GENERIC, db, 'del', key, 1)
        if dest is None:
            client.reply_multi_bulk((key, elem))
        else:
            client.reply_bulk(elem)

    def _range_values(self, value, start, end):
        start = int(start)
        end = int(end)
        if value is not None:
            if start < 0:
                start = len(value) + start
            if end < 0:
                end = len(value) + end + 1
            else:
                end += 1
        return start, end

    def _close_transaction(self, client):
        client.transaction = None
        client.watched_keys = None
        client.flag &= ~self.DIRTY_CAS
        self._watching.discard(client)

    def _flat_info(self):
        info = self._server.info()
        info['server']['redis_version'] = self.version
        e = self._encode_info_value
        for k, values in info.items():
            if isinstance(values, dict):
                yield '#%s' % k
                for key, value in values.items():
                    if isinstance(value, (list, tuple)):
                        value = ', '.join((e(v) for v in value))
                    elif isinstance(value, dict):
                        value = ', '.join(('%s=%s' % (k, e(v))
                                           for k, v in value.items()))
                    else:
                        value = e(value)
                    yield '%s:%s' % (key, value)

    def _get_config(self, name):
        return b''

    def _set_config(self, name, value):
        pass

    def _encode_info_value(self, value):
        return str(value).replace('=',
                                  ' ').replace(',',
                                               ' ').replace('\n', ' - ')

    def _hincrby(self, client, request, N, type):
        check_input(request, N != 3)
        key, field = request[1], request[2]
        try:
            increment = type(request[3])
        except Exception:
            return client.reply_error(
                'value is not an %s or out of range' % type.__name__)
        db = client.db
        hash = db.get(key)
        if hash is None:
            hash = self.hash_type()
            db._data[key] = hash
        elif not isinstance(hash, self.hash_type):
            return client.reply_wrongtype()
        if field in hash:
            try:
                value = type(hash[field])
            except Exception:
                return client.reply_error(
                    'hash value is not an %s' % type.__name__)
            increment += value
        hash[field] = increment
        self._signal(self.NOTIFY_HASH, db, request[0], key, 1)
        return increment

    def _setoper(self, client, oper, keys, dest=None):
        db = client.db
        result = None
        for key in keys:
            value = db.get(key)
            if value is None:
                value = set()
            elif not isinstance(value, set):
                return client.reply_wrongtype()
            if result is None:
                result = value
            else:
                result = getattr(result, oper)(value)
        if dest is not None:
            db.pop(dest)
            if result:
                db._data[dest] = result
                client.reply_int(len(result))
            else:
                client.reply_zero()
        else:
            client.reply_multi_bulk(result)

    def _zsetoper(self, client, request, N):
        check_input(request, N < 3)
        db = client.db
        cmnd = request[0]
        try:
            des = request[1]
            try:
                numkeys = int(request[2])
            except Exception:
                numkeys = 0
            if numkeys <= 0:
                raise ValueError('at least 1 input key is needed for '
                                 'ZUNIONSTORE/ZINTERSTORE')
            sets = []
            for key in request[3:3+numkeys]:
                value = db.get(key)
                if value is None:
                    value = self.zset_type()
                elif not isinstance(value, self.zset_type):
                    return client.reply_wrongtype()
                sets.append(value)
            if len(sets) != numkeys:
                raise ValueError('numkeys does not match number of sets')
            op = set((b'weights', b'aggregate'))
            request = request[3+numkeys:]
            weights = None
            aggregate = sum
            while request:
                name = request[0].lower()
                if name in op:
                    op.discard(name)
                    if name == b'weights':
                        weights = [float(v) for v in request[1:1+numkeys]]
                        request = request[1+numkeys:]
                    elif len(request) > 1:
                        aggregate = self.zset_aggregate.get(request[1])
                        request = request[2:]
                else:
                    raise ValueError(self.SYNTAX_ERROR)
            if not aggregate:
                raise ValueError(self.SYNTAX_ERRO)
            if weights is None:
                weights = [1]*numkeys
            elif len(weights) != numkeys:
                raise ValueError(self.SYNTAX_ERROR)
        except Exception as e:
            return client.reply_error(str(e))
        if cmnd == b'zunionstore':
            result = self.zset_type.union(sets, weights, aggregate)
        else:
            result = self.zset_type.inter(sets, weights, aggregate)
        if db.pop(des) is not None:
            self._signal(self.NOTIFY_GENERIC, db, 'del', des, 1)
        db._data[des] = result
        self._signal(self.NOTIFY_ZSET, db, cmnd, des, len(result))
        client.reply_int(len(result))

    def _score_values(self, min_value, max_value):
        include_min = include_max = True
        if min_value and _ord(min_value[0]) == 40:
            include_min = False
            min_value = min_value[1:]
        if max_value and _ord(max_value[0]) == 40:
            include_max = False
            max_value = max_value[1:]
        return float(min_value), include_min, float(max_value), include_max

    def _info(self):
        keyspace = {}
        stats = {'keyspace_hits': self._hit_keys,
                 'keyspace_misses': self._missed_keys,
                 'expired_keys': self._expired_keys,
                 'keys_changed': self._dirty,
                 'pubsub_channels': len(self._channels),
                 'pubsub_patterns': len(self._patterns),
                 'blocked_clients': self._bpop_blocked_clients}
        persistance = {'rdb_changes_since_last_save': self._dirty,
                       'rdb_last_save_time': self._last_save}
        for db in self.databases.values():
            if len(db):
                keyspace[str(db)] = db.info()
        return {'keyspace': keyspace,
                'stats': stats,
                'persistance': persistance}

    def _client_list(self, client):
        for client in client._producer._concurrent_connections:
            yield ' '.join(self._client_info(client))

    def _client_info(self, client):
        yield 'addr=%s:%s' % client._transport.get_extra_info('addr')
        yield 'fd=%s' % client._transport._sock_fd
        yield 'age=%s' % int(time.time() - client.started)
        yield 'db=%s' % client.database
        yield 'sub=%s' % len(client.channels)
        yield 'psub=%s' % len(client.patterns)
        yield 'cmd=%s' % client.last_command

    def _save(self, async=True):
        writer = self._writer
        if writer and writer.is_alive():
            self.logger.warning('Cannot save, background saving in progress')
        else:
            from multiprocessing import Process
            data = self._dbs()
            self._dirty = 0
            self._last_save = int(time.time())
            if async:
                self.logger.debug('Saving database in background process')
                self._writer = Process(target=save_data,
                                       args=(self.cfg, self._filename, data))
                self._writer.start()
            else:
                self.logger.debug('Saving database')
                save_data(self.cfg, self._filename, data)

    def _dbs(self):
        data = [(db._num, db._data) for db in self.databases.values()
                if len(db._data)]
        return (1, data)

    def _loaddb(self):
        filename = self._filename
        if os.path.isfile(filename):
            self.logger.info('loading data from "%s"', filename)
            with open(filename, 'rb') as file:
                data = pickle.load(file)
            version, dbs = data
            for num, data in dbs:
                db = self.databases.get(num)
                if db is not None:
                    db._data = data

    def _signal(self, type, db, command, key=None, dirty=0):
        self._dirty += dirty
        self._event_handlers[type](db, key, COMMANDS_INFO[command])

    def _publish_clients(self, msg, clients):
        remove = set()
        count = 0
        for client in clients:
            try:
                client._transport.write(msg)
                count += 1
            except Exception:
                remove.add(client)
        if remove:
            clients.difference_update(remove)
        return count

    # EVENT HANDLERS
    def _modified_key(self, key):
        for client in self._watching:
            if key is None or key in client.watched_keys:
                client.flag |= self.DIRTY_CAS

    def _generic_event(self, db, key, command):
        if command.write:
            self._modified_key(key)

    _string_event = _generic_event
    _set_event = _generic_event
    _hash_event = _generic_event
    _zset_event = _generic_event

    def _list_event(self, db, key, command):
        if command.write:
            self._modified_key(key)
        # the key is blocking clients
        if key in db._blocking_keys:
            if key in db._data:
                value = db._data[key]
            elif key in self._expires:
                value = db._expires[key]
            else:
                value = None
            for client in db._blocking_keys.pop(key):
                client.blocked.unblock(client, key, value)

    def _remove_connection(self, client, _, **kw):
        # Remove a client from the server
        self._monitors.discard(client)
        self._watching.discard(client)
        for channel, clients in list(self._channels.items()):
            clients.discard(client)
            if not clients:
                self._channels.pop(channel)
        for pattern, p in list(self._patterns.items()):
            p.clients.discard(client)
            if not p.clients:
                self._patterns.pop(pattern)

    def _write_to_monitors(self, client, request):
        # addr = '%s:%s' % self._transport.get_extra_info('addr')
        cmds = b'" "'.join(request)
        message = '+%s [0 %s] "'.encode('utf-8') + cmds + b'"\r\n'
        remove = set()
        for m in self._monitors:
            try:
                m._transport.write(message)
            except Exception:
                remove.add(m)
        if remove:
            self._monitors.difference_update(remove)

    def _eval_script(self, client, script, request):
        try:
            numkeys = int(request[2])
            if numkeys > 0:
                keys = request[3:3+numkeys]
                if len(keys) < numkeys:
                    raise ValueError
            elif numkeys < 0:
                raise ValueError
            else:
                keys = []
        except Exception:
            return client.reply_error(self.SYNTAX_ERROR)
        args = request[3+numkeys:]
        self.lua.set_global('KEYS', keys)
        self.lua.set_global('ARGV', args)
        result = self.lua.execute(script)
        if isinstance(result, dict):
            if len(result) == 1:
                key = tuple(result)[0]
                keyu = key.upper()
                if keyu == b'OK':
                    return client.reply_ok()
                elif keyu == b'ERR':
                    return client.reply_error(result[key])
                else:
                    return client.reply_multi_bulk(())
            else:
                array = []
                index = 0
                while result:
                    index += 1
                    value = result.pop(index, None)
                    if value is None:
                        break
                    array.append(value)
                result = array
        if isinstance(result, list):
            client.reply_multi_bulk(result)
        elif result is True:
            client.reply_one()
        elif result is False:
            client.reply_bulk()
        elif isinstance(result, (int, float)):
            client.reply_int(int(result))
        else:
            client.reply_bulk(result)


class Db(object):
    '''The database.
    '''
    def __init__(self, num, store):
        self.store = store
        self._num = num
        self._loop = store._loop
        self._data = {}
        self._expires = {}
        self._events = {}
        self._blocking_keys = {}

    def __repr__(self):
        return 'db%s' % self._num
    __str__ = __repr__

    def __len__(self):
        return len(self._data) + len(self._expires)

    def __iter__(self):
        return chain(self._data, self._expires)

    # #########################################################################
    # #    INTERNALS
    def flush(self):
        removed = len(self._data)
        self._data.clear()
        [handle.cancel() for handle, _ in self._expires.values()]
        self._expires.clear()
        self.store._signal(self.store.NOTIFY_GENERIC, self, 'flushdb',
                           dirty=removed)

    def get(self, key, default=None):
        if key in self._data:
            self.store._hit_keys += 1
            return self._data[key]
        elif key in self._expires:
            self.store._hit_keys += 1
            return self._expires[key]
        else:
            self.store._missed_keys += 1
            return default

    def exists(self, key):
        return key in self._data or key in self._expires

    def expire(self, key, timeout):
        if key in self._expires:
            handle, value = self._expires.pop(key)
            handle.cancel()
        elif key in self._data:
            value = self._data.pop(key)
        else:
            return False
        self._expires[key] = (self._loop.call_later(
            timeout, self._do_expire, key), value)
        return True

    def persist(self, key):
        if key in self._expires:
            self.store._hit_keys += 1
            handle, value = self._expires.pop(key)
            handle.cancel()
            self._data[key] = value
            return True
        elif key in self._data:
            self.store._hit_keys += 1
        else:
            self.store._missed_keys += 1
        return False

    def ttl(self, key, m=1):
        if key in self._expires:
            self.store._hit_keys += 1
            handle, value = self._expires[key]
            return max(0, int(m*(handle._when - self._loop.time())))
        elif key in self._data:
            self.store._hit_keys += 1
            return -1
        else:
            self.store._missed_keys += 1
            return -2

    def info(self):
        return {'Keys': len(self._data),
                'expires': len(self._expires)}

    def pop(self, key, value=None):
        if not value:
            if key in self._data:
                value = self._data.pop(key)
                return value
            elif key in self._expires:
                handle, value = self._expires.pop(key)
                handle.cancel()
                return value

    def rem(self, key):
        if key in self._data:
            self.store._hit_keys += 1
            self._data.pop(key)
            self.store._signal(self.store.NOTIFY_GENERIC, self, 'del', key, 1)
            return 1
        elif key in self._expires:
            self.store._hit_keys += 1
            handle, _ = self._expires.pop(key)
            handle.cancel()
            self.store._signal(self.store.NOTIFY_GENERIC, self, 'del', key, 1)
            return 1
        else:
            self.store._missed_keys += 1
            return 0

    def _do_expire(self, key):
        if key in self._expires:
            handle, value, = self._expires.pop(key)
            handle.cancel()
            self.store._expired_keys += 1



#
# Required by Cython to build Hiredis extensions
#
import os
import sys
from distutils.core import setup
from distutils.extension import Extension
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)

from Cython.Distutils import build_ext
from Cython.Build import cythonize

include_dirs = []
ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
   # 2.6's distutils.msvc9compiler can raise an IOError when failing to
   # find the compiler
   ext_errors += (IOError,)

class BuildFailed(Exception):

    def __init__(self, msg=None):
        if not msg:
            msg = str(sys.exc_info()[1])
        self.msg = msg


class tolerant_build_ext(build_ext):
    # This class allows C extension building to fail. From SQLAlchemy

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildFailed

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            raise BuildFailed
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(sys.exc_info()[1]): # works with both py 2/3
                raise BuildFailed
            raise

lib_path = 'extensions'


def lib_extension():
    path = os.path.join(lib_path, 'lib')
    include_dirs.append(path)
    return Extension('pulsar.utils.lib',
                     [os.path.join(path, 'lib.pyx')],
                     include_dirs=include_dirs)


def libparams():
    extensions = [lib_extension()]
    return {'ext_modules': cythonize(extensions),
            'cmdclass': {'build_ext' : tolerant_build_ext},
            'include_dirs': include_dirs}



'''The example is a :ref:`WSGI application <apps-wsgi>`
with a :ref:`websocket middleware <apps-ws>` which connects to an
IMAP4 server to retrieve and send emails.

The connection with the IMAP4 server is obtained using the IMAP4 API in
twisted 12.3 or later. The example uses
:ref:`pulsar-twisted integration <tutorials-twisted>` module.

To run the server you need to create a :mod:`config.py` file in the
the :mod:`examples.webmail` directory containing::

    # the adress of your mail server
    mail_incoming ='ssl:host=imap.gmail.com:port=993'
    # mail_username & mail_password
    mail_username=
    mail_password=

And type::

    python manage.py

Open a web browser at http://localhost:8060 and you should see the web app.

For information on twisted IMAP4 client library check this example:

http://twistedmatrix.com/documents/current/mail/examples/imap4client.py

Other python examples of webmail:

* https://github.com/khamidou/kite

Implementation
==================

.. autoclass:: WsMail
   :members:
   :member-order: bysource

'''
import os
import sys

from pulsar import coroutine_return
from pulsar.apps import ws, wsgi
from pulsar.utils.log import process_global
from pulsar.utils.system import json
try:
    from pulsar.apps.tx import twisted
    from twisted.internet import protocol, endpoints, reactor
    from twisted.mail import imap4
except ImportError:  # pragma    nocover
    twisted = None    # This is for when we build docs

ASSET_DIR = os.path.join(os.path.dirname(__file__), 'assets')


def mail_client(cfg, timeout=10):
    '''Create a new mail client using twisted IMAP4 library.'''
    key = (cfg.mail_incoming, cfg.mail_username, cfg.mail_password)
    client = process_global(key)
    if not client:
        endpoint = endpoints.clientFromString(reactor, cfg.mail_incoming)
        endpoint._timeout = timeout
        factory = protocol.Factory()
        factory.protocol = imap4.IMAP4Client
        client = yield endpoint.connect(factory)
        yield client.login(cfg.mail_username, cfg.mail_password)
        yield client.select('INBOX')
        process_global(key, client, True)
    coroutine_return(client)
    # info = yield client.fetchEnvelope(imap4.MessageSet(1))
    # print 'First message subject:', info[1]['ENVELOPE'][1]


class WsMail(ws.WS):
    '''A :class:`.WS` handler for fetching and sending mail via the
    twisted IMAP4 library
    '''
    def on_open(self, websocket):
        '''When the websocket starts, it create a new mail client.'''
        request = websocket.handshake
        client = yield mail_client(request.cfg)
        # add the mail client to the environ cache
        request.cache.mailclient = client
        # retrieve the list of mailboxes and them to the client
        yield self._send_mailboxes(websocket)

    def on_message(self, websocket, msg):
        request = websocket.request
        client = request.cache.mailclient
        if msg and client:
            msg = json.loads(msg)
            if 'mailbox' in msg:
                mailbox = yield client.examine(msg['mailbox'])
                self.write(request, json.dumps({'mailbox': mailbox}))

    def _send_mailboxes(self, websocket):
        request = websocket.handshake
        result = yield request.cache.mailclient.list("", "*")
        result = sorted([e[2] for e in result])
        websocket.write(json.dumps({'list': result}))


class WebMail(wsgi.LazyWsgi):

    def setup(self, environ):
        return wsgi.WsgiHandler([ws.WebSocket('/message', WsMail()),
                                 wsgi.MediaRouter('/media', ASSET_DIR),
                                 wsgi.Router('/', get=self.home)])

    def home(self, request):
        data = open(os.path.join(ASSET_DIR, 'mail.html')).read()
        response = request.response
        request.response.content = data % request.environ
        request.response.content_type = 'text/html'
        return response


def server(name='webmail', callable=None, **kwargs):
    return wsgi.WSGIServer(name=name, callable=WebMail(), **kwargs)


if __name__ == '__main__':  # pragma nocover
    server().start()



import pulsar
from pulsar import coroutine_return, get_application
from pulsar.apps import rpc

from .backend import Task, TaskNotAvailable


__all__ = ['TaskQueueRpcMixin']


def task_to_json(task):
    if task:
        if isinstance(task, (list, tuple)):
            task = [task_to_json(t) for t in task]
        elif isinstance(task, Task):
            task = task.to_json()
    return task


class TaskQueueRpcMixin(rpc.JSONRPC):
    '''A :class:`.JSONRPC` mixin for communicating with  a :class:`.TaskQueue`.

    To use it, you need to have an :ref:`RPC application <apps-rpc>`
    and a :ref:`task queue <apps-taskqueue>` application installed in the
    :class:`.Arbiter`.

    :parameter taskqueue: instance or name of the :class:`.TaskQueue`
        application which exposes the remote procedure calls.

    '''
    _task_backend = None

    def __init__(self, taskqueue, **kwargs):
        if not isinstance(taskqueue, str):
            taskqueue = taskqueue.name
        self.taskqueue = taskqueue
        super(TaskQueueRpcMixin, self).__init__(**kwargs)

    ########################################################################
    #    REMOTES
    def rpc_job_list(self, request, jobnames=None):
        '''Return the list of Jobs registered with task queue with meta
        information.

        If a list of ``jobnames`` is given, it returns only jobs
        included in the list.
        '''
        task_backend = yield self.task_backend()
        coroutine_return(task_backend.job_list(jobnames=jobnames))

    def rpc_next_scheduled_tasks(self, request, jobnames=None):
        return self._rq(request, 'next_scheduled', jobnames=jobnames)

    def rpc_queue_task(self, request, jobname=None, **kw):
        '''Queue a new ``jobname`` in the task queue.

        The task can be of any type as long as it is registered in the
        task queue registry. To check the available tasks call the
        :meth:`rpc_job_list` function.

        It returns the task :attr:`~Task.id`.
        '''
        result = yield self.queue_task(request, jobname, **kw)
        coroutine_return(task_to_json(result))

    def rpc_get_task(self, request, id=None):
        '''Retrieve a task from its id'''
        if id:
            task_backend = yield self.task_backend()
            result = yield task_backend.get_task(id)
            coroutine_return(task_to_json(result))

    def rpc_get_tasks(self, request, **filters):
        '''Retrieve a list of tasks which satisfy key-valued filters'''
        if filters:
            task_backend = yield self.task_backend()
            result = yield task_backend.get_tasks(**filters)
            coroutine_return(task_to_json(result))

    def rpc_wait_for_task(self, request, id=None, timeout=None):
        '''Wait for a task to have finished.

        :param id: the id of the task to wait for.
        :param timeout: optional timeout in seconds.
        :return: the json representation of the task once it has finished.
        '''
        if id:
            task_backend = yield self.task_backend()
            result = yield task_backend.wait_for_task(id, timeout=timeout)
            coroutine_return(task_to_json(result))

    def rpc_num_tasks(self, request):
        '''Return the approximate number of tasks in the task queue.'''
        task_backend = yield self.task_backend()
        coroutine_return(task_backend.num_tasks())

    ########################################################################
    #    INTERNALS
    def task_backend(self):
        if not self._task_backend:
            app = yield get_application(self.taskqueue)
            self._task_backend = app.get_backend()
        coroutine_return(self._task_backend)

    def queue_task(self, request, jobname, meta_data=None, **kw):
        if not jobname:
            raise rpc.InvalidParams('"jobname" is not specified!')
        meta_data = meta_data or {}
        meta_data.update(self.task_request_parameters(request))
        task_backend = yield self.task_backend()
        result = yield task_backend.queue_task(jobname, meta_data, **kw)
        coroutine_return(result)

    def task_request_parameters(self, request):
        '''**Internal function** which returns a dictionary of parameters
        to be passed to the :class:`.Task` class constructor.

        This function can be overridden to add information about
        the type of request, who made the request and so forth.
        It must return a dictionary.
        By default it returns an empty dictionary.'''
        return {}

    def _rq(self, request, action, *args, **kw):
        return pulsar.send(self.taskqueue, action, *args, **kw)



import os
from struct import pack, unpack

cimport cython
from common cimport websocket_mask


cdef class Frame:
    cdef int _opcode
    cdef bint _final
    cdef int _payload_length
    cdef bytes _masking_key
    cdef object _body

    def __cinit__(self, int opcode, bint final, int payload_length):
        self._opcode = opcode
        self._final = final
        self._payload_length = payload_length

    @property
    def final(self):
        return self._final

    @property
    def opcode(self):
        return self._opcode

    @property
    def body(self):
        return self._body

    @property
    def masking_key(self):
        return self._masking_key


cdef class FrameParser:
    '''Encoder and decoder of WebSocket frames.

    * kind = 0, server (decode masked messages, encode unmasked messages)
    * kind = 1, client (decode unmasked messages, encode masked messages)
    * kind = 2, unmasked both encoding and decoding
    * kind = 3, masked both encoding and decoding
    '''
    cdef Frame frame
    cdef object buffer
    cdef object ProtocolError
    cdef int kind
    cdef int _version
    cdef int _decode_mask_length
    cdef int _encode_mask_length
    cdef tuple _opcodes
    cdef object _extensions
    cdef object _protocols

    def __cinit__(self, int version, int kind, object ProtocolError,
                  extensions=None, protocols=None):
        self._version = version
        self.kind = kind
        self.frame = None
        self.buffer = bytearray()
        self.ProtocolError = ProtocolError
        self._opcodes = (0, 1, 2, 8, 9, 10)
        self._encode_mask_length = 0
        self._decode_mask_length = 0
        if kind == 0:
            self._decode_mask_length = 4
        elif kind == 1:
            self._encode_mask_length = 4
        elif kind == 3:
            self._decode_mask_length = 4
            self._encode_mask_length = 4
        self._extensions = extensions
        self._protocols = protocols

    @property
    def version(self):
        return self._version

    @property
    def max_payload(self):
        return 1 << 63

    @property
    def decode_mask_length(self):
        return self._decode_mask_length

    @property
    def encode_mask_length(self):
        return self._encode_mask_length

    def ping(self, body=None):
        '''return a `ping` :class:`Frame`.'''
        return self.encode(body, opcode=9)

    def pong(self, body=None):
        '''return a `pong` :class:`Frame`.'''
        return self.encode(body, opcode=10)

    def close(self, body=None):
        '''return a `close` :class:`Frame`.'''
        return self.encode(body, opcode=8)

    def continuation(self, body, final=True):
        '''return a `continuation` :class:`Frame`.'''
        return self.encode(body, opcode=0, final=final)

    def encode(self, message, final=True, bytes masking_key=None,
               int opcode=-1, int rsv1=0, int rsv2=0, int rsv3=0):
        '''Encode a ``message`` for writing into the wire.

        The message length cannot exceed :attr:`max_payload`
        To produce several frames for a given large message use
        :meth:`multi_encode'.
        '''
        cdef bytes data
        cdef int fin = 1 if final else 0
        opcode, masking_key, data = self._info(message, opcode, masking_key)
        return self._encode(data, opcode, masking_key, fin,
                            rsv1, rsv2, rsv3)

    def multi_encode(self, message, bytes masking_key=None, int opcode=-1,
                     int rsv1=0, int rsv2=0, int rsv3=0,
                     cython.ulonglong max_payload=0):
        '''Encode a ``message`` into several frames depending on size.

        Returns a generator of bytes to be sent over the wire.
        '''
        cdef bytes data
        cdef bytes chunk
        cdef int fin
        max_payload = max(2, max_payload or 1 << 63)
        opcode, masking_key, data = self._info(message, opcode, masking_key)
        #
        while data:
            if len(data) >= max_payload:
                chunk, data, fin = (data[:max_payload],
                                    data[max_payload:], 0)
            else:
                chunk, data, fin = data, b'', 1
            yield self._encode(chunk, opcode, masking_key, fin,
                               rsv1, rsv2, rsv3)

    def decode(self, bytes data=None):
        cdef int fin, rsv1, rsv2, rsv3, opcode, payload_length
        cdef Frame frame = self.frame
        cdef int mask_length = self._decode_mask_length
        cdef bytes chunk
        #
        if data:
            self.buffer.extend(data)
        if frame is None:
            if len(self.buffer) < 2:
                return
            first_byte, second_byte = unpack("BB", self.buffer[:2])
            fin = (first_byte >> 7) & 1
            rsv1 = (first_byte >> 6) & 1
            rsv2 = (first_byte >> 5) & 1
            rsv3 = (first_byte >> 4) & 1
            opcode = first_byte & 0xf
            if fin not in (0, 1):
                raise self.ProtocolError('FIN must be 0 or 1')
            if bool(mask_length) != bool(second_byte & 0x80):
                if mask_length:
                    raise self.ProtocolError('unmasked client frame.')
                else:
                    raise self.ProtocolError('masked server frame.')
            payload_length = second_byte & 0x7f
            # All control frames MUST have a payload length of 125 bytes
            # or less
            if opcode > 7:
                if payload_length > 125:
                    raise self.ProtocolError(
                        'WEBSOCKET control frame too large')
                elif not fin:
                    raise self.ProtocolError(
                        'WEBSOCKET control frame fragmented')
            chunk = self._chunk(2)
            self.frame = frame = Frame(opcode, <bint>fin, payload_length)

        if frame._masking_key is None:
            if frame._payload_length == 126:
                if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
                    return
                chunk = self._chunk(2)
                frame._payload_length = unpack("!H", chunk)[0]
            elif frame._payload_length == 127:
                if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask
                    return
                chunk = self._chunk(8)
                frame._payload_length = unpack("!Q", chunk)[0]
            elif len(self.buffer) < mask_length:
                return
            if mask_length:
                frame._masking_key = self._chunk(mask_length)
            else:
                frame._masking_key = b''

        if len(self.buffer) >= frame._payload_length:
            self.frame = None
            chunk = self._chunk(frame._payload_length)
            if self._extensions:
                for extension in self._extensions:
                    chunk = extension.receive(frame, self.buffer)
            if frame._masking_key:
                chunk = websocket_mask(chunk, frame._masking_key,
                                       len(chunk), len(frame._masking_key))
            if frame.opcode == 1:
                frame._body = chunk.decode("utf-8", "replace")
            else:
                frame._body = chunk
            return frame

    cdef bytes _encode(self, bytes data, int opcode, bytes masking_key,
                       int fin, int rsv1, int rsv2, int rsv3):
        cdef object buffer = bytearray()
        cdef cython.longlong length = len(data)
        cdef int mask_bit = 128 if masking_key else 0

        buffer.append(((fin << 7) | (rsv1 << 6) | (rsv2 << 5) |
                       (rsv3 << 4) | opcode))

        if length < 126:
            buffer.append(mask_bit | length)
        elif length < 65536:
            buffer.append(mask_bit | 126)
            buffer.extend(pack('!H', length))
        elif length < (1 << 63):
            buffer.append(mask_bit | 127)
            buffer.extend(pack('!Q', length))
        else:
            raise self.ProtocolError('WEBSOCKET frame too large')
        if masking_key:
            buffer.extend(masking_key)
            buffer.extend(websocket_mask(data, masking_key,
                                         length, len(masking_key)))
        else:
            buffer.extend(data)
        return bytes(buffer)

    cdef tuple _info(self, message, int opcode, bytes masking_key):
        cdef int mask_length = self._encode_mask_length
        #
        if mask_length:
            masking_key = masking_key or os.urandom(4)
            assert len(masking_key) == mask_length, "bad masking key"
        else:
            masking_key = b''
        if opcode == -1:
            opcode = 1 if isinstance(message, string_type) else 2
        data = to_bytes(message or b'', 'utf-8')
        if opcode not in self._opcodes:
            raise self.ProtocolError('WEBSOCKET opcode a reserved value')
        elif opcode > 7:
            if len(data) > 125:
                raise self.ProtocolError('WEBSOCKET control frame too large')
            if opcode == 8:
                #TODO CHECK CLOSE FRAME STATUS CODE
                pass
        return opcode, masking_key, data

    cdef bytes _chunk(self, int length):
        cdef bytes chunk = bytes(self.buffer[:length])
        self.buffer = self.buffer[length:]
        return chunk



try:
    import signal
except ImportError:
    signal = None


__all__ = ['ALL_SIGNALS',
           'SIG_NAMES',
           'SKIP_SIGNALS',
           'MAXFD',
           'set_proctitle',
           'get_proctitle']


SIG_NAMES = {}
MAXFD = 1024
SKIP_SIGNALS = frozenset(('KILL', 'STOP', 'WINCH'))


def all_signals():
    if signal:
        for sig in dir(signal):
            if sig.startswith('SIG') and sig[3] != "_":
                val = getattr(signal, sig)
                if isinstance(val, int):
                    name = sig[3:]
                    if name not in SKIP_SIGNALS:
                        SIG_NAMES[val] = name
                        yield name


ALL_SIGNALS = tuple(all_signals())


try:
    from setproctitle import setproctitle, getproctitle

    def set_proctitle(title):
        setproctitle(title)
        return True

    def get_proctitle():
        return getproctitle()

except ImportError:  # pragma    nocover

    def set_proctitle(title):
        return False

    def get_proctitle():
        pass



import sys
from functools import partial
import logging

from pulsar import AsyncObject, multi_async, task, coroutine_return
from pulsar.utils.system import json
from pulsar.utils.structures import AttributeDictionary
from pulsar.utils.security import gen_unique_id
from pulsar.utils.pep import range
from pulsar.utils.tools import checkarity
from pulsar.apps.wsgi import Json
from pulsar.apps.http import HttpClient
from pulsar.utils.httpurl import JSON_CONTENT_TYPES

from .handlers import RpcHandler, InvalidRequest, exception


__all__ = ['JSONRPC', 'JsonProxy']


logger = logging.getLogger('pulsar.jsonrpc')


class JSONRPC(RpcHandler):
    '''An :class:`.RpcHandler` for JSON-RPC services.

    Design to comply with the `JSON-RPC 2.0`_ Specification.

    JSON-RPC is a lightweight remote procedure call protocol
    designed to be simple.
    A remote method is invoked by sending a request to a remote service,
    the request is a single object serialised using JSON.

    .. _`JSON-RPC 2.0`: http://www.jsonrpc.org/specification
    '''
    version = '2.0'

    def __call__(self, request):
        return Json(self._call(request)).http_response(request)

    def _call(self, request):
        response = request.response
        data = {}
        exc_info = None
        callable = None
        try:
            try:
                data = yield request.body_data()
            except ValueError:
                raise InvalidRequest(
                    status=415, msg='Content-Type must be application/json')
            if data.get('jsonrpc') != self.version:
                raise InvalidRequest(
                    'jsonrpc must be supplied and equal to "%s"' %
                    self.version)
            params = data.get('params')
            if isinstance(params, dict):
                args, kwargs = (), params
            else:
                args, kwargs = tuple(params or ()), {}
            #
            callable = self.get_handler(data.get('method'))
            result = yield callable(request, *args, **kwargs)
        except Exception as exc:
            result = exc
            exc_info = sys.exc_info()
        #
        res = {'id': data.get('id'), "jsonrpc": self.version}
        if exc_info:
            code = getattr(result, 'fault_code', -32603)
            msg = None
            if isinstance(result, TypeError) and callable:
                msg = checkarity(callable, args, kwargs, discount=1)
            msg = msg or str(result) or 'JSON RPC exception'
            if code == -32603:
                logger.error(msg, exc_info=exc_info)
            else:
                logger.warning(msg)
            error = {'code': code,
                     'message': msg,
                     'data': getattr(result, 'data', '')}
            response.status_code = getattr(result, 'status', 500)
            res['error'] = error
        else:
            res['result'] = result
        coroutine_return(res)


class JsonCall:
    slots = ('_client', '_name')

    def __init__(self, client, name):
        self._client = client
        self._name = name

    def __repr__(self):
        return self._name
    __str__ = __repr__

    @property
    def url(self):
        return self._client.url

    @property
    def name(self):
        return self._name

    def __getattr__(self, name):
        name = "%s%s%s" % (self._name, self._client.separator, name)
        return self.__class__(self._client, name)

    def __call__(self, *args, **kwargs):
        return self._client._call(self._name, *args, **kwargs)


class JsonProxy(AsyncObject):
    '''A python Proxy class for :class:`.JSONRPC` Servers.

    :param url: server location
    :param version: JSON-RPC server version. Default ``2.0``
    :param id: optional request id, generated if not provided.
        Default ``None``.
    :param data: Extra data to include in all requests. Default ``None``.
    :param full_response: return the full Http response rather than
        just the content.
    :param http: optional http client. If provided it must have the ``request``
        method available which must be of the form::

            http.request(url, body=..., method=...)

        Default ``None``.

    Lets say your RPC server is running at ``http://domain.name.com/``::

        >>> a = JsonProxy('http://domain.name.com/')
        >>> a.add(3,4)
        7
        >>> a.ping()
        'pong'

    '''
    separator = '.'
    default_version = '2.0'
    default_timeout = 30

    def __init__(self, url, version=None, data=None,
                 full_response=False, http=None, timeout=None, **kw):
        self._url = url
        self._version = version or self.__class__.default_version
        self._full_response = full_response
        self._data = data if data is not None else {}
        if not http:
            timeout = timeout if timeout is not None else self.default_timeout
            http = HttpClient(timeout=timeout, **kw)
        http.headers['accept'] = 'application/json, text/*; q=0.5'
        http.headers['content-type'] = 'application/json'
        self._http = http

    @property
    def url(self):
        return self._url

    @property
    def version(self):
        return self._version

    @property
    def _loop(self):
        return self._http._loop

    def makeid(self):
        '''Can be re-implemented by your own Proxy'''
        return gen_unique_id()

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__url)

    def __str__(self):
        return self.__repr__()

    def __getattr__(self, name):
        return JsonCall(self, name)

    @task
    def _call(self, name, *args, **kwargs):
        data = self._get_data(name, *args, **kwargs)
        body = json.dumps(data).encode('utf-8')
        resp = yield self._http.post(self._url, data=body)
        if self._full_response:
            coroutine_return(resp)
        else:
            content = resp.decode_content()
            if resp.is_error:
                if 'error' not in content:
                    resp.raise_for_status()
            coroutine_return(self.loads(content))

    def _get_data(self, func_name, *args, **kwargs):
        id = self.makeid()
        params = self.get_params(*args, **kwargs)
        data = {'method': func_name, 'params': params, 'id': id,
                'jsonrpc': self._version}
        return data

    def get_params(self, *args, **kwargs):
        '''
        Create an array or positional or named parameters
        Mixing positional and named parameters in one
        call is not possible.
        '''
        kwargs.update(self._data)
        if args and kwargs:
            raise ValueError('Cannot mix positional and named parameters')
        if args:
            return list(args)
        else:
            return kwargs

    def loads(self, obj):
        if isinstance(obj, dict):
            if 'error' in obj:
                error = obj['error']
                raise exception(error.get('code'), error.get('message'))
            else:
                return obj.get('result')
        return obj



from .server import PulsarDS, DEFAULT_PULSAR_STORE_ADDRESS, pulsards_url
from .client import COMMANDS_INFO, redis_to_py_pattern
from .parser import (PyRedisParser, RedisParser, redis_parser, ResponseError,
                     InvalidResponse, NoScriptError)



'''A parser for redis messages
'''
import sys
from itertools import starmap

ispy3k = sys.version_info >= (3, 0)
if ispy3k:
    long = int
    string_type = str
else:   # pragma    nocover
    string_type = unicode

nil = b'$-1\r\n'
null_array = b'*-1\r\n'

REPLAY_TYPE = frozenset((b'$',   # REDIS_REPLY_STRING,
                         b'*',   # REDIS_REPLY_ARRAY,
                         b':',   # REDIS_REPLY_INTEGER,
                         b'+',   # REDIS_REPLY_STATUS,
                         b'-'))  # REDIS_REPLY_ERROR


class String(object):
    __slots__ = ('_length', 'next')

    def __init__(self, length, next):
        self._length = length
        self.next = next

    def decode(self, parser, result):
        parser._current = None
        length = self._length
        if length >= 0:
            b = parser._inbuffer
            if len(b) >= length+2:
                parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
                if parser.encoding:
                    return chunk.decode(parser.encoding)
                else:
                    return chunk
            else:
                parser._current = self
                return False


class ArrayTask(object):
    __slots__ = ('_length', '_response', 'next')

    def __init__(self, length, next):
        self._length = length
        self._response = []
        self.next = next

    def decode(self, parser, result):
        parser._current = None
        length = self._length
        if length >= 0:
            response = self._response
            if result is not False:
                response.append(result)
            while len(response) < length:
                result = parser._get(self)
                if result is False:
                    break
                response.append(result)
            if len(response) == length:
                parser._current = None
                return response
            elif not parser._current:
                parser._current = self
            return False


class Parser(object):
    '''A python parser for redis.'''
    encoding = None

    def __init__(self, protocolError, responseError):
        self.protocolError = protocolError
        self.responseError = responseError
        self._current = None
        self._inbuffer = bytearray()

    def on_connect(self, connection):
        if connection.decode_responses:
            self.encoding = connection.encoding

    def on_disconnect(self):
        pass

    def feed(self, buffer):
        '''Feed new data into the buffer'''
        self._inbuffer.extend(buffer)

    def get(self):
        '''Called by the protocol consumer'''
        if self._current:
            return self._resume(self._current, False)
        else:
            return self._get(None)

    def bulk(self, value):
        if value is None:
            return nil
        else:
            return ('$%d\r\n' % len(value)).encode('utf-8') + value + b'\r\n'

    def multi_bulk_len(self, len):
        return ('*%s\r\n' % len).encode('utf-8')

    def multi_bulk(self, args):
        '''Multi bulk encoding for list/tuple ``args``
        '''
        return null_array if args is None else b''.join(self._pack(args))

    def pack_command(self, args):
        '''Encode a command to send to the server.

        Used by redis clients
        '''
        return b''.join(self._pack_command(args))

    def pack_pipeline(self, commands):
        '''Packs pipeline commands into bytes.'''
        pack = lambda *args: b''.join(self._pack_command(args))
        return b''.join(starmap(pack, (args for args, _ in commands)))

    #    INTERNALS
    def _pack_command(self, args):
        crlf = b'\r\n'
        yield ('*%d\r\n' % len(args)).encode('utf-8')
        for value in args:
            if isinstance(value, string_type):
                value = value.encode('utf-8')
            elif not isinstance(value, bytes):
                value = str(value).encode('utf-8')
            yield ('$%d\r\n' % len(value)).encode('utf-8')
            yield value
            yield crlf

    def _pack(self, args):
        crlf = b'\r\n'
        yield ('*%d\r\n' % len(args)).encode('utf-8')
        for value in args:
            if value is None:
                yield nil
            elif isinstance(value, bytes):
                yield ('$%d\r\n' % len(value)).encode('utf-8')
                yield value
                yield crlf
            elif isinstance(value, string_type):
                value = value.encode('utf-8')
                yield ('$%d\r\n' % len(value)).encode('utf-8')
                yield value
                yield crlf
            elif hasattr(value, 'items'):
                for value in self._pack(tuple(self._lua_dict(value))):
                    yield value
            elif hasattr(value, '__len__'):
                for value in self._pack(value):
                    yield value
            else:
                value = str(value).encode('utf-8')
                yield ('$%d\r\n' % len(value)).encode('utf-8')
                yield value
                yield crlf

    def _lua_dict(self, d):
        index = 0
        while True:
            index += 1
            v = d.get(index)
            if v is None:
                break
            yield v

    def _get(self, next):
        b = self._inbuffer
        length = b.find(b'\r\n')
        if length >= 0:
            self._inbuffer, response = b[length+2:], bytes(b[:length])
            rtype, response = response[:1], response[1:]
            if rtype == b'-':
                return self.responseError(response.decode('utf-8'))
            elif rtype == b':':
                return long(response)
            elif rtype == b'+':
                return response
            elif rtype == b'$':
                task = String(long(response), next)
                return task.decode(self, False)
            elif rtype == b'*':
                task = ArrayTask(long(response), next)
                return task.decode(self, False)
            else:
                # Clear the buffer and raise
                self._inbuffer = bytearray()
                raise self.protocolError('Protocol Error')
        else:
            return False

    def buffer(self):
        '''Current buffer'''
        return bytes(self._inbuffer)

    def _resume(self, task, result):
        result = task.decode(self, result)
        if result is not False and task.next:
            return self._resume(task.next, result)
        else:
            return result



'''Tests the RPC "calculator" example.'''
import unittest

from pulsar import send, new_event_loop, coroutine_return
from pulsar.apps import rpc
from pulsar.apps.test import dont_run_with_thread

from .manage import server, Root, Calculator


class TestRpcOnThread(unittest.TestCase):
    app_cfg = None
    concurrency = 'thread'
    # used for both keep-alive and timeout in JsonProxy
    # long enough to allow to wait for tasks
    rpc_timeout = 500

    @classmethod
    def setUpClass(cls):
        name = 'calc_' + cls.concurrency
        s = server(bind='127.0.0.1:0', name=name, concurrency=cls.concurrency)
        cls.app_cfg = yield send('arbiter', 'run', s)
        cls.uri = 'http://{0}:{1}'.format(*cls.app_cfg.addresses[0])
        cls.p = rpc.JsonProxy(cls.uri, timeout=cls.rpc_timeout)
        cls.sync = rpc.JsonProxy(cls.uri, loop=new_event_loop())

    @classmethod
    def tearDownClass(cls):
        if cls.app_cfg:
            return send('arbiter', 'kill_actor', cls.app_cfg.name)

    def setUp(self):
        self.assertEqual(self.p.url, self.uri)
        self.assertTrue(str(self.p))
        proxy = self.p.bla
        self.assertEqual(proxy.name, 'bla')
        self.assertEqual(proxy.url, self.uri)
        self.assertEqual(proxy._client, self.p)
        self.assertEqual(str(proxy), 'bla')

    def test_wsgi_handler(self):
        cfg = self.app_cfg
        self.assertTrue(cfg.callable)
        wsgi_handler = cfg.callable.handler({})
        self.assertEqual(len(wsgi_handler.middleware), 1)
        router = wsgi_handler.middleware[0]
        self.assertEqual(router.route.path, '/')
        root = router.post
        self.assertEqual(len(root.subHandlers), 1)
        hnd = root.subHandlers['calc']
        self.assertFalse(hnd.isroot())
        self.assertEqual(hnd.subHandlers, {})

    # Pulsar server commands
    def test_ping(self):
        response = yield self.p.ping()
        self.assertEqual(response, 'pong')

    def test_functions_list(self):
        result = yield self.p.functions_list()
        self.assertTrue(result)
        d = dict(result)
        self.assertTrue('ping' in d)
        self.assertTrue('echo' in d)
        self.assertTrue('functions_list' in d)
        self.assertTrue('calc.add' in d)
        self.assertTrue('calc.divide' in d)

    def test_time_it(self):
        '''Ping server 5 times'''
        bench = yield self.p.timeit('ping', 5)
        self.assertTrue(len(bench.result), 5)
        self.assertTrue(bench.taken)

    # Test Object method
    def test_check_request(self):
        result = yield self.p.check_request('check_request')
        self.assertTrue(result)

    def test_add(self):
        response = yield self.p.calc.add(3, 7)
        self.assertEqual(response, 10)

    def test_subtract(self):
        response = yield self.p.calc.subtract(546, 46)
        self.assertEqual(response, 500)

    def test_multiply(self):
        response = yield self.p.calc.multiply(3, 9)
        self.assertEqual(response, 27)

    def test_divide(self):
        response = yield self.p.calc.divide(50, 25)
        self.assertEqual(response, 2)

    def test_info(self):
        response = yield self.p.server_info()
        self.assertTrue('server' in response)
        server = response['server']
        self.assertTrue('version' in server)
        app = response['monitors'][self.app_cfg.name]
        if self.concurrency == 'thread':
            self.assertFalse(app['workers'])
            worker = app
        else:
            workers = app['workers']
            self.assertEqual(len(workers), 1)
            worker = workers[0]
        name = '%sserver' % self.app_cfg.name
        if name in worker:
            self._check_tcpserver(worker[name]['server'])

    def _check_tcpserver(self, server):
        sockets = server['sockets']
        if sockets:
            self.assertEqual(len(sockets), 1)
            sock = sockets[0]
            self.assertEqual(sock['address'],
                             '%s:%s' % self.app_cfg.addresses[0])

    def test_invalid_params(self):
        self.async.assertRaises(rpc.InvalidParams, self.p.calc.add, 50, 25, 67)

    def test_invalid_params_fromApi(self):
        self.async.assertRaises(rpc.InvalidParams, self.p.calc.divide,
                                50, 25, 67)

    def test_invalid_function(self):
        p = self.p
        yield self.async.assertRaises(rpc.NoSuchFunction, p.foo, 'ciao')
        yield self.async.assertRaises(rpc.NoSuchFunction,
                                      p.blabla)
        yield self.async.assertRaises(rpc.NoSuchFunction,
                                      p.blabla.foofoo)
        yield self.async.assertRaises(rpc.NoSuchFunction,
                                      p.blabla.foofoo.sjdcbjcb)

    def testInternalError(self):
        self.async.assertRaises(rpc.InternalError, self.p.calc.divide,
                                'ciao', 'bo')

    def testCouldNotserialize(self):
        self.async.assertRaises(rpc.InternalError, self.p.dodgy_method)

    def testpaths(self):
        '''Fetch a sizable ammount of data'''
        response = yield self.p.calc.randompaths(num_paths=20, size=100,
                                                 mu=1, sigma=2)
        self.assertTrue(response)

    def test_echo(self):
        response = yield self.p.echo('testing echo')
        self.assertEqual(response, 'testing echo')

    # Synchronous client
    def test_sync_ping(self):
        self.assertEqual(self.sync.ping(), 'pong')
        self.assertEqual(self.sync.ping(), 'pong')

    def test_docs(self):
        handler = Root({'calc': Calculator})
        self.assertEqual(handler.parent, None)
        self.assertEqual(handler.root, handler)
        self.assertRaises(rpc.NoSuchFunction, handler.get_handler,
                          'cdscsdcscd')
        calc = handler.subHandlers['calc']
        self.assertEqual(calc.parent, handler)
        self.assertEqual(calc.root, handler)
        docs = handler.docs()
        self.assertTrue(docs)
        response = yield self.p.documentation()
        self.assertEqual(response, docs)


@dont_run_with_thread
class TestRpcOnProcess(TestRpcOnThread):
    concurrency = 'process'



'''Test twisted integration'''
import unittest

import pulsar
from pulsar import multi_async
from pulsar.utils.pep import to_bytes, to_string
from pulsar.utils.security import gen_unique_id

from examples.echo.manage import server, EchoProtocol

try:
    # This import must be done before importing twisted
    from pulsar.apps.tx import twisted
    from twisted.internet.protocol import Factory, Protocol
    from twisted.internet.defer import Deferred
    from twisted.internet.endpoints import TCP4ClientEndpoint
    from twisted.internet import reactor

    class EchoClient(Protocol):
        '''Twisted client to the Echo server in the examples.echo module'''
        separator = EchoProtocol.separator
        connected = False

        def __init__(self):
            self.buffer = b''
            self.requests = {}

        def connectionMade(self):
            self.connected = True

        def __call__(self, msg):
            id = to_bytes(gen_unique_id()[:8])
            self.requests[id] = d = Deferred()
            self.transport.write(id + to_bytes(msg) + self.separator)
            return d

        def dataReceived(self, data):
            sep = self.separator
            idx = data.find(sep)
            if idx >= 0:
                if self.buffer:
                    msg = self.buffer + msg
                    self.buffer = b''
                id, msg, data = data[:8], data[8:idx], data[idx+len(sep):]
                d = self.requests.pop(id)
                d.callback(to_string(msg))
                if data:
                    self.dataReceived(data)
            else:
                self.buffer += data

    class EchoClientFactory(Factory):
        protocol = EchoClient

    def get_client(address):
        point = TCP4ClientEndpoint(reactor, *address)
        return point.connect(EchoClientFactory())

except ImportError:
    twisted = None


@unittest.skipUnless(twisted, 'Requires twisted')
class TestTwistedIntegration(unittest.TestCase):
    server_cfg = None

    @classmethod
    def setUpClass(cls):
        s = server(name=cls.__name__.lower(), bind='127.0.0.1:0')
        cls.server_cfg = yield pulsar.send('arbiter', 'run', s)
        cls.address = cls.server_cfg.addresses[0]

    @classmethod
    def tearDownClass(cls):
        if cls.server_cfg:
            return pulsar.send('arbiter', 'kill_actor', cls.server_cfg.name)

    def test_echo_client(self):
        client = yield get_client(self.address)
        self.assertTrue(client.connected)
        result = yield client('Hello')
        self.assertEqual(result, 'Hello')
        result = yield client('Ciao')
        self.assertEqual(result, 'Ciao')

    def test_multi_requests(self):
        client = yield get_client(self.address)
        results = yield multi_async((client('Msg%s' % n) for n in range(20)))
        self.assertEqual(len(results), 20)
        for n, result in enumerate(results):
            self.assertEqual(result, 'Msg%s' % n)


@unittest.skipUnless(twisted, 'Requires twisted')
class TestPulsarReactor(unittest.TestCase):

    def test_meta(self):
        self.assertTrue(reactor.running)
        self.assertEqual(reactor.threadpool, None)
        self.assertEqual(reactor.waker, None)

    def test_switched_off_methods(self):
        self.assertRaises(NotImplementedError, reactor.spawnProcess)



from collections import deque
from functools import partial

from pulsar.utils.pep import iteritems

from .futures import (Future, maybe_async, InvalidStateError,
                      future_result_exc, AsyncObject)


__all__ = ['EventHandler', 'Event', 'OneTime']


class AbstractEvent(AsyncObject):
    '''Abstract event handler.'''
    _silenced = False
    _handlers = None
    _fired = 0

    @property
    def silenced(self):
        '''Boolean indicating if this event is silenced.

        To silence an event one uses the :meth:`silence` method.
        '''
        return self._silenced

    @property
    def handlers(self):
        if self._handlers is None:
            self._handlers = []
        return self._handlers

    def bind(self, callback):
        '''Bind a ``callback`` to this event.
        '''
        self.handlers.append(callback)

    def fired(self):
        '''The number of times this event has fired'''
        return self._fired

    def fire(self, arg, **kwargs):
        '''Fire this event.'''
        raise NotImplementedError

    def clear(self):
        if self._handlers:
            self._handlers[:] = []

    def silence(self):
        '''Silence this event.

        A silenced event won't fire when the :meth:`fire` method is called.
        '''
        self._silenced = True


class Event(AbstractEvent):
    '''The default implementation of :class:`AbstractEvent`.
    '''
    def __init__(self, loop=None):
        self._loop = loop

    def __repr__(self):
        return repr(self._handlers)
    __str__ = __repr__

    def fire(self, arg, **kwargs):
        if not self._silenced:
            self._fired += self._fired + 1
            if self._handlers:
                for hnd in self._handlers:
                    try:
                        maybe_async(hnd(arg, **kwargs), self._loop)
                    except Exception:
                        self.logger.exception('Exception while firing event')
        return self


class OneTime(Future, AbstractEvent):
    '''An :class:`AbstractEvent` which can be fired once only.

    This event handler is a subclass of :class:`.Future`.
    Implemented mainly for the one time events of the :class:`EventHandler`.
    '''
    @property
    def handlers(self):
        if self._handlers is None:
            self._handlers = deque()
        return self._handlers

    def bind(self, callback):
        '''Bind a ``callback`` to this event.
        '''
        if not self.done():
            self.handlers.append(callback)
        else:
            result, exc = future_result_exc(self)
            self._loop.call_soon(
                lambda: maybe_async(callback(result, exc=exc), self._loop))

    def fire(self, arg, exc=None, **kwargs):
        '''The callback handlers registered via the :meth:~AbstractEvent.bind`
        method are executed first.

        :param arg: the argument
        :param exc: optional exception
        '''
        if not self._silenced:
            if self._fired:
                raise InvalidStateError('already fired')
            self._fired = 1
            self._process(arg, exc, kwargs)
        return self

    def clear(self):
        if self._handlers:
            self._handlers.clear()

    def _process(self, arg, exc, kwargs, future=None):
        while self._handlers:
            hnd = self._handlers.popleft()
            try:
                result = maybe_async(hnd(arg, exc=exc, **kwargs), self._loop)
            except Exception:
                self.logger.exception('Exception while firing event')
            else:
                if isinstance(result, Future):
                    result.add_done_callback(
                        partial(self._process, arg, exc, kwargs))
                    return
        if exc:
            self.set_exception(exc)
        else:
            self.set_result(arg)


class EventHandler(AsyncObject):
    '''A Mixin for handling events.

    It handles :class:`OneTime` events and :class:`Event` that occur
    several times.
    '''
    ONE_TIME_EVENTS = ()
    '''Event names which occur once only.'''
    MANY_TIMES_EVENTS = ()
    '''Event names which occur several times.'''
    def __init__(self, loop=None, one_time_events=None,
                 many_times_events=None):
        one = self.ONE_TIME_EVENTS
        if one_time_events:
            one = set(one)
            one.update(one_time_events)
        events = dict(((name, OneTime(loop=loop)) for name in one))
        many = self.MANY_TIMES_EVENTS
        if many_times_events:
            many = set(many)
            many.update(many_times_events)
        events.update(((name, Event(loop=loop)) for name in many))
        self._events = events

    @property
    def events(self):
        '''The dictionary of all events.
        '''
        return self._events

    def event(self, name):
        '''Returns the :class:`Event` at ``name``.

        If no event is registered for ``name`` returns nothing.
        '''
        event = self._events.get(name)
        if event:
            assert self._loop, "No event loop for %s" % self
            event._loop = self._loop
        return event

    def fired_event(self, name):
        event = self._events.get(name)
        return event._fired if event else 0

    def bind_event(self, name, callback):
        '''Register a ``callback`` with ``event``.

        The callback must be a callable accepting one positional parameter
        and at least the ``exc`` optional parameter::

            def callback(arg, ext=None):
                ...

            o.bind_event('start', callback)

        the instance firing the event or the first positional argument
        passed to the :meth:`fire_event` method.

        :param name: the event name. If the event is not available a warning
            message is logged.
        :param callback: a callable receiving one positional parameter. It
            can also be a list/tuple of callables.
        :return: nothing.
        '''
        if name not in self._events:
            self._events[name] = Event()
        event = self._events[name]
        event.bind(callback)

    def bind_events(self, **events):
        '''Register all known events found in ``events`` key-valued parameters.

        The events callbacks can be specified as a single callable or as
        list/tuple of callabacks or (callback, erroback) tuples.
        '''
        for name in self._events:
            if name in events:
                callbacks = events[name]
                self.bind_event(name, events[name])

    def fire_event(self, name, *args, **kwargs):
        """Dispatches ``arg`` or ``self`` to event ``name`` listeners.

        * If event at ``name`` is a one-time event, it makes sure that it was
          not fired before.

        :param args: optional argument passed as positional parameter to the
            event handler.
        :param kwargs: optional key-valued parameters to pass to the event
            handler. Can only be used for
            :ref:`many times events <many-times-event>`.
        :return: the :class:`Event` fired
        """
        if not args:
            arg = self
        elif len(args) == 1:
            arg = args[0]
        else:
            raise TypeError('fire_event expected at most 1 argument got %s' %
                            len(args))
        event = self.event(name)
        if event:
            try:
                event.fire(arg, **kwargs)
            except InvalidStateError:
                self.logger.error('Event %s already fired' % name)
            return event
        else:
            self.logger.warning('Unknown event "%s" for %s', name, self)

    def silence_event(self, name):
        '''Silence event ``name``.

        This causes the event not to fire at the :meth:`fire_event` method
        is invoked with the event ``name``.
        '''
        event = self._events.get(name)
        if event:
            event.silence()

    def copy_many_times_events(self, other):
        '''Copy :ref:`many times events <many-times-event>` from  ``other``.

        All many times events of ``other`` are copied to this handler
        provided the events handlers already exist.
        '''
        if isinstance(other, EventHandler):
            events = self._events
            for name, event in iteritems(other._events):
                if isinstance(event, Event) and event._handlers:
                    ev = events.get(name)
                    # If the event is available add it
                    if ev:
                        for callback in event._handlers:
                            ev.bind(callback)



'''
Handle basic and digest authentication on the server.

HttpAuthenticate
~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: HttpAuthenticate
   :members:
   :member-order: bysource



Authorization header parser
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autofunction:: parse_authorization_header
'''
import os
import time
from base64 import b64decode

from pulsar import HttpException
from pulsar.utils.httpurl import (parse_dict_header, hexmd5, hexsha1,
                                  quote_header_value, DEFAULT_CHARSET)
from pulsar.utils.pep import iteritems, to_bytes


__all__ = ['HttpAuthenticate', 'parse_authorization_header']


_require_quoting = frozenset(['domain', 'nonce', 'opaque', 'realm'])


class HttpAuthenticate(HttpException):
    '''Exception when ``basic`` or ``digest`` authentication is required.

    This HttpException is raised with status code ``401`` and the extra
    ``WWW_Authenticate`` header if ``type`` is either ``basic`` or ``digest``.
    '''
    def __init__(self, type, realm=None, **options):
        realm = realm or 'authentication required'
        if type == 'basic':
            value = self._auth_header(type, realm=realm)
        elif type == 'digest':
            value = self.digest_auth_header(realm, **options)
        else:
            value = None
        if value:
            value = [('WWW-Authenticate', value)]
        super(HttpAuthenticate, self).__init__(status=401, headers=value)

    def digest_auth_header(self, realm=None, nonce=None, qop=None, opaque=None,
                           algorithm=None, stale=None):
        options = {}
        if nonce is None:
            nonce = hexmd5(to_bytes('%d' % time.time()) + os.urandom(10))
            if opaque is None:
                opaque = hexmd5(os.urandom(10))
        if stale:
            options['stale'] = 'TRUE'
        if opaque is not None:
            options['opaque'] = opaque
        if algorithm is not None:
            options['algorithm'] = algorithm
        if qop is None:
            qop = ('auth',)
        return self._auth_header('digest', realm=realm, nonce=nonce,
                                 qop=', '.join(qop), **options)

    def _auth_header(self, type, **options):
        """Convert the stored values into a WWW-Authenticate header."""
        return '%s %s' % (type.title(), ', '.join((
            '%s=%s' % (key, quote_header_value(
                value, allow_token=key not in _require_quoting))
            for key, value in iteritems(options)
        )))


class BasicAuth(object):
    def __init__(self, username, password):
        self.username = username
        self.password = password

    def authenticated(self, environ, username=None, password=None, **params):
        return username == self.username and password == self.password

    def __repr__(self):
        return 'Basic: %s' % self.username
    __str__ = __repr__


class DigestAuth(object):
    def __init__(self, username, password=None, options=None):
        self.username = username
        self.password = password
        self.last_nonce = None
        self.options = options or {}
        self.algorithm = self.options.pop('algorithm', 'MD5')

    def __repr__(self):
        return 'Digest: %s' % self.username
    __str__ = __repr__

    def authenticated(self, environ, username=None, password=None, **params):
        '''Called by the server to check if client is authenticated.'''
        if username != self.username:
            return False
        o = self.options
        qop = o.get('qop')
        method = environ['REQUEST_METHOD']
        uri = environ.get('PATH_INFO', '')
        ha1 = self.ha1(o['realm'], password)
        ha2 = self.ha2(qop, method, uri)
        if qop is None:
            response = hexmd5(":".join((ha1, self.nonce, ha2)))
        elif qop == 'auth' or qop == 'auth-int':
            response = hexmd5(":".join((ha1, o['nonce'], o['nc'],
                                        o['cnonce'], qop, ha2)))
        else:
            raise ValueError("qop value are wrong")
        return o['response'] == response

    def hex(self, x):
        if self.algorithm == 'MD5':
            return hexmd5(x)
        elif self.algorithm == 'SHA1':
            return hexsha1(x)
        else:
            raise ValueError('Unknown algorithm %s' % self.algorithm)

    def ha1(self, realm, password):
        return self.hex('%s:%s:%s' % (self.username, realm, password))

    def ha2(self, qop, method, uri, body=None):
        if qop == "auth" or qop is None:
            return self.hex("%s:%s" % (method, uri))
        elif qop == "auth-int":
            return self.hex("%s:%s:%s" % (method, uri, self.hex(body)))
        raise ValueError()


digest_parameters = frozenset(('username', 'realm', 'nonce', 'uri', 'nc',
                               'cnonce', 'response'))


def parse_authorization_header(value, charset='utf-8'):
    '''Parse an HTTP basic/digest authorisation header.

    :param value: the authorisation header to parse.
    :return: either `None` if the header was invalid or
        not given, otherwise an :class:`Auth` object.
    '''
    if not value:
        return
    try:
        auth_type, auth_info = value.split(None, 1)
        auth_type = auth_type.lower()
    except ValueError:
        return
    if auth_type == 'basic':
        try:
            up = b64decode(auth_info.encode(DEFAULT_CHARSET)).decode(charset)
            username, password = up.split(':', 1)
        except Exception:
            return
        return BasicAuth(username, password)
    elif auth_type == 'digest':
        auth_map = parse_dict_header(auth_info)
        if not digest_parameters.difference(auth_map):
            return DigestAuth(auth_map.pop('username'), options=auth_map)
